# Copyright (c) 2018 - 2021 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from decimal import Decimal
from gzip import GzipFile
import os
from pathlib import Path
import shutil
import sys
import tarfile
from tarfile import TarFile


class ArInfo(object):
    """Represents a single member in an ar archive."""
    HEADER_SIZE = 60

    mtime = 0
    uid = 0
    gid = 0
    perm = 0

    def __init__(self, fileobj, offset, size, name=None):
        self.fileobj = fileobj
        self.offset = offset
        self.size = size
        self.name = name
        self.pos = 0
        self.mode = 'rb'
        self.padded_size = size + (size % 2)

    def tell(self):
        return self.pos

    def seek(self, offset, whence=os.SEEK_SET):
        if whence == os.SEEK_SET:
            new_pos = offset
        elif whence == os.SEEK_CUR:
            new_pos = self.pos + offset
        elif whence == os.SEEK_END:
            new_pos = self.size + offset

        self.pos = max(0, min(new_pos, self.size))

    def read(self, size=None):
        max_size = self.size - self.pos
        if size is None:
            size = max_size
        else:
            size = min(size, max_size)

        self.fileobj.seek(self.offset + self.pos)
        self.pos += size

        buf = self.fileobj.read(size)
        if len(buf) != size:
            raise IOError("unexpected end of data")
        return buf

    def write(self, buf):
        if self.mode != 'ab':
            raise IOError(f"bad operation for mode {self.mode!r}")

        self.fileobj.seek(self.offset + self.pos)
        self.fileobj.write(buf)
        self.size += max(0, len(buf) - (self.size - self.pos))
        self.pos += len(buf)
        return len(buf)

    def close(self):
        if self.mode == 'rb':
            return

        remainder = self.size % 2
        self.padded_size = self.size + remainder
        if remainder > 0:
            self.fileobj.seek(self.offset + self.size)
            self.fileobj.write(b'\n' * remainder)

        self.fileobj.seek(self.offset - self.HEADER_SIZE)
        self.fileobj.write(self.tobuf())
        self.arfile.offset = self.offset + self.padded_size
        self.mode = 'rb'

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if type is None:
            self.close()
        else:
            # An exception occurred. Don't call close() because we cannot afford to try writing the header
            self.mode = 'rb'

    @classmethod
    def frombuf(cls, fileobj, buf, data_offset):
        if len(buf) != cls.HEADER_SIZE:
            raise IOError(f"Too short a header for ar file: {len(buf)} instead of {cls.HEADER_SIZE}")
        member_name, mtime, uid, gid, perm, size, _ = (
            buf[ 0:16],  # noqa: E201
            buf[16:28],
            buf[28:34],
            buf[34:40],
            buf[40:48],
            buf[48:58],
            buf[58:60],
        )
        member_name = member_name.rstrip(b' ').rstrip(b'/').decode('ASCII')

        try:
            size = int(size)
        except ValueError:
            raise IOError("Non-numeric file size in ar header")

        arinfo = cls(fileobj, data_offset, size, Path(member_name) if member_name else None)

        arinfo.mtime = int(mtime)
        arinfo.uid   = int(uid)
        arinfo.gid   = int(gid)
        arinfo.perm  = int(perm, 8)

        return arinfo

    def tobuf(self):
        buf = f"{self.name!s:<16.16}{self.mtime:<12d}{self.uid:<6d}{self.gid:<6d}{self.perm:<8o}{self.size:<10d}`\n".encode("ASCII")
        if len(buf) != self.HEADER_SIZE:
            raise IOError(f"Exceeding maximum header size: {buf}")
        return buf


class ArFile(object):
    """Provides an interface to ar archives."""
    arinfo = ArInfo

    def __init__(self, name=None, mode='r', fileobj=None):
        modes = {'r': 'rb', 'w': 'wb'}
        if mode not in modes:
            raise ValueError("mode must be 'r' or 'w'")
        self.mode = mode

        if not fileobj:
            fileobj = open(name, modes[mode])
            self._extfileobj = False
        else:
            if name is None:
                name = getattr(fileobj, 'name', None)
            self._extfileobj = True

        try:
            self.name = os.path.abspath(name) if name else None
            self.fileobj = fileobj
            self.closed = False

            if mode == 'r':
                self.read_signature = False
            if mode == 'w':
                self.fileobj.write(b'!<arch>\n')
            self.offset = self.fileobj.tell()
        except:  # noqa: E722: we re-raise, so it's not a problem
            if not self._extfileobj:
                fileobj.close()
            self.closed = True
            raise

    def close(self):
        if self.closed:
            return

        self.closed = True
        if not self._extfileobj:
            self.fileobj.close()

    def next(self):
        if self.closed:
            raise IOError("ArFile is closed")
        if self.mode != 'r':
            raise IOError(f"bad operation for mode {self.mode!r}")

        self.fileobj.seek(self.offset)
        if not self.read_signature:
            signature = self.fileobj.read(8)
            self.offset += len(signature)
            expected_signature = b'!<arch>\n'
            if len(signature) < len(expected_signature):
                raise StopIteration
            if signature != expected_signature:
                raise IOError('Invalid ar file signature')
            self.read_signature = True

        file_header = self.fileobj.read(self.arinfo.HEADER_SIZE)
        self.offset += len(file_header)
        if len(file_header) < self.arinfo.HEADER_SIZE:
            raise StopIteration
        arinfo = self.arinfo.frombuf(self.fileobj, file_header, self.offset)
        self.offset += arinfo.padded_size
        return arinfo

    def __next__(self):
        return self.next()

    def __iter__(self):
        if self.closed:
            raise IOError("ArFile is closed")
        if self.mode != 'r':
            raise IOError(f"bad operation for mode {self.mode!r}")

        self.offset = 0
        self.read_signature = False
        return self

    def appendfile(self, arinfo):
        if self.closed:
            raise IOError("ArFile is closed")
        if self.mode != 'w':
            raise IOError(f"bad operation for mode {self.mode!r}")

        arinfo = copy.copy(arinfo)

        self.fileobj.seek(self.offset)
        buf = arinfo.tobuf()
        self.fileobj.write(buf)
        arinfo.offset = self.offset + len(buf)

        arinfo.fileobj = self.fileobj
        arinfo.arfile = self
        arinfo.mode = 'ab'
        arinfo.size = 0
        arinfo.pos = 0

        return arinfo

    def addfile(self, arinfo, fileobj):
        with self.appendfile(arinfo) as outfile:
            shutil.copyfileobj(fileobj, outfile)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


def normalize(filename, fileobj=None, outname='', outfileobj=None, source_date_epoch=0):
    """Make the given file as close to reproducible as possible. Mostly be clamping timestamps to source_date_epoch."""
    if (fileobj is None or outfileobj is None) and not os.path.isfile(filename):
        return

    if filename.suffix == ".tar" or filename.suffixes[-2:] == [".tar", ".gz"]:
        if outfileobj is None:
            archivefile = outfile = open(filename.with_suffix(filename.suffix + ".tmp"), 'wb')
        else:
            archivefile = outfile = outfileobj
        with TarFile.open(filename, fileobj=fileobj) as in_archive:
            try:
                compress = False
                if filename.suffix == ".gz":
                    compress = True
                    archivefile = GzipFile(filename=outname, mode='wb', compresslevel=9, fileobj=outfile, mtime=source_date_epoch)

                with TarFile.open(outname, fileobj=archivefile, format=tarfile.PAX_FORMAT, mode='w', encoding='UTF-8') as out_archive:
                    if sys.version_info < (3, 9, 0):
                        # Erase major/minor fields for non-device files as other tar-archive producing tools do and Python does since 3.9.0.

                        archive_props = {
                            "offset": 0,
                            "next_header_offset": 0,
                        }
                        original_write = archivefile.write

                        def write_archivefile(data):
                            # Intercept writes to the archive file and rewrite any headers matching our criteria
                            for block_idx in range((len(data) + tarfile.BLOCKSIZE - 1) // tarfile.BLOCKSIZE):
                                block = data[block_idx * tarfile.BLOCKSIZE : (block_idx + 1) * tarfile.BLOCKSIZE]
                                cur_offset = archive_props["offset"]
                                archive_props["offset"] += len(block)

                                if cur_offset != archive_props["next_header_offset"]:
                                    continue

                                try:
                                    tarinfo = out_archive.tarinfo.frombuf(block, out_archive.encoding, out_archive.errors)
                                except tarfile.EOFHeaderError:  # type: ignore[attr-defined]
                                    break

                                block_count = (tarinfo.size + tarfile.BLOCKSIZE - 1) // tarfile.BLOCKSIZE
                                archive_props["next_header_offset"] = archive_props["offset"] + block_count * tarfile.BLOCKSIZE

                                if tarinfo.devmajor != 0 or tarinfo.devmajor != 0 or tarinfo.type in (tarfile.CHRTYPE, tarfile.BLKTYPE):
                                    continue

                                # Erase major/minor fields for non-device files as other tar-archive producing tools do.
                                if not isinstance(data, bytearray):
                                    data = bytearray(data)

                                DEVNUMBERS_OFFSET = 329
                                DEVNUMBERS_LENGTH = 2 * 8
                                DEVNUMBERS_RANGE = slice(
                                    block_idx * tarfile.BLOCKSIZE + DEVNUMBERS_OFFSET,
                                    block_idx * tarfile.BLOCKSIZE + DEVNUMBERS_OFFSET + DEVNUMBERS_LENGTH,
                                )
                                CHECKSUM_OFFSET = 148
                                CHECKSUM_LENGTH = 8
                                CHECKSUM_RANGE = slice(
                                    block_idx * tarfile.BLOCKSIZE + CHECKSUM_OFFSET,
                                    block_idx * tarfile.BLOCKSIZE + CHECKSUM_OFFSET + CHECKSUM_LENGTH,
                                )

                                data[DEVNUMBERS_RANGE] = b"\0" * DEVNUMBERS_LENGTH

                                # Recompute checksum, but whipe checksum field (with spaces) before doing so
                                data[CHECKSUM_RANGE] = b" " * CHECKSUM_LENGTH
                                chksum, *_ = tarfile.calc_chksums(  # type: ignore[attr-defined]
                                    data[block_idx * tarfile.BLOCKSIZE : (block_idx + 1) * tarfile.BLOCKSIZE]
                                )
                                data[CHECKSUM_RANGE] = b"%06o\0 " % (chksum,)

                            return original_write(data)

                        archivefile.write = write_archivefile

                    # Sorting the file list ensures that we don't depend on the order that files appear on disk
                    for member in sorted(in_archive, key=lambda x: x.name):
                        # Clamping mtime to source_date_epoch ensures that source files are the only sources of timestamps, not build time
                        mtime = min(Decimal(member.pax_headers.pop("mtime", member.mtime)), source_date_epoch)
                        # Store in PAX header if it needs sub-integer precision
                        if int(mtime) != mtime:
                            member.mtime = int(mtime)
                            member.pax_headers["mtime"] = f"{mtime:f}"
                        else:
                            member.mtime = int(mtime)
                            member.pax_headers.pop("mtime", None)

                        # Don't store atime or ctime. These are just too volatile.
                        member.pax_headers.pop("atime", None)
                        member.pax_headers.pop("ctime", None)

                        # Prevent including the account details of the account used to execute the build
                        if member.uid == os.getuid() or member.gid == os.getgid():
                            member.uid = 0
                            member.gid = 0
                        member.uname = ''
                        member.gname = ''

                        fileobj = (in_archive.extractfile(member) if member.isfile() else None)
                        out_archive.addfile(member, fileobj)
                if compress:
                    archivefile.close()
            finally:
                if outfileobj is not None:
                    outfile.close()
        if outfileobj is None:
            os.utime(outfile.name, (source_date_epoch, source_date_epoch))
            os.rename(outfile.name, filename)

    elif filename.suffix == ".deb":
        with ArFile(filename) as in_pkg, ArFile(filename.with_suffix(filename.suffix + ".tmp"), 'w') as out_pkg:
            # A valid Debian package contains these files in this order
            expected_files = [
                (u'debian-binary',),
                (u'control.tar', u'control.tar.gz', u'control.tar.xz'),
                (u'data.tar', u'data.tar.gz', u'data.tar.bz2', u'data.tar.xz'),
            ]
            for pkg_member in in_pkg:
                if expected_files:
                    expected = expected_files.pop(0)
                    if str(pkg_member.name) not in expected:
                        break

                # Clamping mtime to source_date_epoch ensures that source files are the only sources of timestamps, not build time
                pkg_member.mtime = min(pkg_member.mtime, source_date_epoch)

                # Prevent including permission information
                pkg_member.uid = 0
                pkg_member.gid = 0
                pkg_member.perm = 0o100644

                with out_pkg.appendfile(pkg_member) as outfile:
                    normalize(pkg_member.name, fileobj=pkg_member, outname=pkg_member.name, outfileobj=outfile, source_date_epoch=source_date_epoch)
            else:
                in_pkg.close()
                out_pkg.close()
                os.utime(out_pkg.name, (source_date_epoch, source_date_epoch))
                os.rename(out_pkg.name, in_pkg.name)

    elif fileobj is not None and outfileobj is not None:
        shutil.copyfileobj(fileobj, outfileobj)



# Copyright (c) 2019 - 2020 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..versioning import CarusoVer


def test_carver_tag_formatting():
    version_str = '7.4.47+PI13.0'
    version = CarusoVer.parse(version_str)
    assert version == CarusoVer(7, 4, 47, (), 13, 0)
    assert str(version) == version_str
    assert version.default_tag_name.format(version=version) == version_str



# Copyright (c) 2020 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import os.path
import sys
from functools import (
    partial,
    wraps,
)
from pathlib import (
    Path,
    PurePath,
)
from typing import (
    AbstractSet,
    Any,
    Callable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from ..compat import metadata

import click_log
import git
import pytest
from click import ClickException
from click.testing import CliRunner
from typeguard import typechecked

from ..cli import utils
from . import (
    hopic_cli,
    sgr_re,
    source_date_epoch,
)

try:
    # Only available from Python >= 3.8 onwards
    from typing import Protocol
except ImportError:
    FileLike = Any
else:
    class FileLike(Protocol):
        name: str

        def read(self) -> str:
            ...


PACKAGE: str = __package__.split('.')[0]
_hopic_version_str = f"{PACKAGE}=={metadata.version(PACKAGE)}"
_root_dir = Path(__file__).parent / '..' / '..'
_example_dir = _root_dir / 'examples'

_git_time = f"{source_date_epoch} +0000"
_author = git.Actor('Bob Tester', 'bob@example.net')
_commitargs = dict(
        author_date=_git_time,
        commit_date=_git_time,
        author=_author,
        committer=_author,
    )


@pytest.fixture
def run_hopic(caplog, monkeypatch, tmp_path):
    @typechecked
    def run_hopic(
        *args: Union[List, Tuple, Callable[[], Any]],
        config: Union[str, FileLike, None] = None,
        files: Optional[Mapping[Union[str, PurePath], Union[str, Tuple[str, Callable[[Union[str, PurePath]], Any]]]]] = None,
        env: Optional[Mapping[str, str]] = None,
        monkeypatch_injector: Callable[[Any], None] = lambda _: None,
        umask: int = 0o022,
        commit_count: int = 0,
        dirty: Union[bool, str] = False,
        tag: Optional[str] = None,
        rundir: Optional[Path] = None,
    ):
        result = None
        commit = None
        runner = CliRunner(mix_stderr=False, env=env)
        umask = os.umask(umask)
        try:
            with monkeypatch.context() as dir_ctx:
                if rundir is None:
                    rundir = tmp_path / "rundir"
                if not rundir.exists():
                    rundir.mkdir(parents=True)
                dir_ctx.chdir(rundir)

                files = ({} if files is None else files.copy())
                if config is not None:
                    try:
                        fn = config.read
                    except AttributeError:
                        fn = lambda: config  # noqa: E731
                    files[getattr(config, "name", ".ci/hopic-ci-config.yaml")] = fn()

                if files:
                    with git.Repo.init() as repo:
                        for fname, content in files.items():
                            on_file_created_callback = lambda path: None  # noqa: E731
                            if not isinstance(content, str):
                                (content, on_file_created_callback) = content
                            if '/' in fname and not os.path.exists(os.path.dirname(fname)):
                                os.makedirs(os.path.dirname(fname))
                            with open(fname, 'w') as f:
                                f.write(content)
                            on_file_created_callback(fname)
                        repo.index.add(files.keys())
                        commit = repo.index.commit(message='Initial commit', **_commitargs)
                        if tag:
                            repo.create_tag(tag)
                        for i in range(commit_count):
                            commit = repo.index.commit(message=f"Some commit {i}", **_commitargs)
                        if dirty:
                            if dirty is True:
                                dirty = "dirty_file"
                            (rundir / dirty).write_text("dirty")
                            repo.index.add((dirty,))  # do not commit to create the dirty state
                else:
                    assert config is None and commit_count == 0 and dirty is False and tag is None

                for arg in args:
                    if callable(arg):
                        arg()
                        continue

                    orig_main = hopic_cli.main
                    rv = []

                    @wraps(orig_main)
                    def mock_main(*args, **kwargs):
                        with monkeypatch.context() as m:
                            # Ensure pytest can capture our logging
                            m.setattr(click_log, "basic_config", lambda: None)
                            m.setattr(logging.getLogger("hopic"), "setLevel", lambda _: None)
                            m.setattr(logging.getLogger("git"), "setLevel", lambda _: None)

                            monkeypatch_injector(m)
                            rv.append(orig_main(*args, **kwargs))
                            return rv[-1]

                    with monkeypatch.context() as call_ctx:
                        call_ctx.setattr(hopic_cli, "main", mock_main)
                        result = runner.invoke(hopic_cli, [str(a) for a in arg], standalone_mode=False)

                    if result.stdout_bytes:
                        print(result.stdout, end='')
                    if result.stderr_bytes:
                        print(result.stderr, end='', file=sys.stderr)

                    if result.exception is not None and not isinstance(result.exception, (SystemExit, ClickException)):
                        raise result.exception
                    if isinstance(result.exception, ClickException):
                        result.exit_code = result.exception.exit_code
                    if result.exit_code == 0 and isinstance(rv[-1], int):
                        result.exit_code = rv[-1]

                    result.commit = commit
                    result.logs = tuple(
                        (rec.levelno, sgr_re.sub("", rec.getMessage())) for rec in caplog.records if rec.name == "hopic" or rec.name.startswith("hopic.")
                    )
                    yield result

                    if result.exit_code != 0:
                        return
        finally:
            os.umask(umask)

    run_hopic.toprepo = tmp_path / "repo"
    # Make _all_ logging levels available for capture by pytest
    caplog.set_level("DEBUG", logger="git")
    caplog.set_level("DEBUG", logger="hopic")
    return run_hopic


@pytest.fixture(autouse=True)
def pip_freeze_constant(monkeypatch):
    """Prevent invoking 'pip freeze' to improve execution speed"""
    with monkeypatch.context() as m:
        m.setattr(utils, 'installed_pkgs', lambda: _hopic_version_str)
        yield m


def _data_file_paths(
    datadir: Union[str, PurePath],
    *,
    recurse: bool = False,
    suffices: Optional[AbstractSet[str]] = None,
):
    for entry in datadir.iterdir():
        if recurse and entry.is_dir():
            yield from _data_file_paths(entry, suffices=suffices)
            continue

        if suffices is not None and entry.suffix not in suffices:
            continue

        if entry.is_dir() or entry.is_socket() or entry.is_block_device():
            continue

        yield entry

    if recurse and datadir == _example_dir:
        from ..cli.utils import determine_config_file_name

        yield determine_config_file_name(None, _root_dir)


def _data_file_path_id(
    datadir: Union[str, PurePath],
    name,
):
    try:
        return os.path.relpath(name, datadir)
    except TypeError:
        return str(name)


def pytest_generate_tests(metafunc):
    for fixture in metafunc.fixturenames:
        if fixture.endswith('example_file'):
            datadir = _example_dir
            dir_prefix = fixture[: -len('_example_file')]
            if dir_prefix:
                datadir = datadir.joinpath(*dir_prefix.split('_'))

            metafunc.parametrize(
                fixture,
                _data_file_paths(datadir, recurse=not dir_prefix, suffices={'.yml', '.yaml'}),
                ids=partial(_data_file_path_id, _example_dir),
            )



# Copyright (c) 2018 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import logging
import os
import shlex
import signal
import stat
import subprocess
import sys
import tempfile
import time
import typing
import urllib.parse
from collections.abc import (
    Mapping,
    Sequence,
)
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
)

import click
from dateutil.tz import tzutc
import git

from . import (
    autocomplete,
    extensions,
)
from .utils import (
    is_publish_branch,
)

from .. import (
    credentials,
    binary_normalize,
)
from ..build import (
    FatalSignal,
    DockerContainers,
    volume_spec_to_docker_param,
    HopicGitInfo,
)
from ..config_reader import (
    CredentialEncoding,
    CredentialType,
    RunOnChange,
    expand_docker_volume_spec,
    expand_vars,
)
from ..errors import (
    MissingCredentialVarError,
    MissingFileError,
    StepTimeoutExpiredError,
    UnknownPhaseError,
)
from ..execution import echo_cmd_click as echo_cmd
from ..git_time import (
    restore_mtime_from_git,
    to_git_time,
)
from ..types import PathLike
from .global_obj import initialize_global_variables_from_config

log = logging.getLogger(__name__)


@click.pass_context
def build_variant(ctx, variant, cmds, hopic_git_info, *, exec_stdout=None, cwd: str = "${WORKSPACE}", config_based_volume_vars={}):
    cfg = ctx.obj.config

    images = cfg['image']
    try:
        image = images[variant]
    except KeyError:
        image = images.get('default', None)

    docker_in_docker = False

    volume_vars = config_based_volume_vars
    volume_vars.update(ctx.obj.volume_vars.copy())

    if hopic_git_info.submit_ref is not None:
        volume_vars['GIT_BRANCH'] = hopic_git_info.submit_ref
    if hopic_git_info.submit_commit is not None:
        # Give commands executing inside a container image a different view than outside
        volume_vars["GIT_COMMIT"] = str(hopic_git_info.submit_commit)
        git_commit_time = hopic_git_info.submit_commit.committed_datetime
        # RFC 3339 formatted (strict subset of ISO 8601)
        volume_vars["GIT_COMMIT_TIME"] = f"{git_commit_time:%Y-%m-%dT%H:%M:%S.%f%z}"
    else:
        git_commit_time = None

    volume_vars["BUILD_NAME"] = os.environ.get("BUILD_NAME", "unknown")
    volume_vars["BUILD_NUMBER"] = os.environ.get("BUILD_NUMBER", "NaN")
    volume_vars["BUILD_URL"] = os.environ.get("BUILD_URL", "")

    # Hack to represent an empty commit list
    volume_vars["SOURCE_COMMITS"] = volume_vars["AUTOSQUASHED_COMMITS"] = "HEAD..HEAD"
    if hopic_git_info.target_commit and hopic_git_info.source_commit:
        volume_vars["AUTOSQUASHED_COMMITS"] = volume_vars["SOURCE_COMMITS"] = f"{hopic_git_info.target_commit}..{hopic_git_info.source_commit}"
    if hopic_git_info.target_commit and hopic_git_info.autosquashed_commit:
        volume_vars["AUTOSQUASHED_COMMITS"] = f"{hopic_git_info.target_commit}..{hopic_git_info.autosquashed_commit}"

    mandatory_artifacts = []
    mandatory_junit = []
    optional_artifacts = []
    worktree_commits: Dict[PathLike, List[str]] = {}
    variant_credentials = {}
    extra_docker_run_args = []
    with DockerContainers() as volumes_from:
        # If the branch is not allowed to publish, skip the publish phase. If run_on_change is set to 'always', phase will be run anyway regardless of
        # this condition. For build phase, run_on_change is set to 'always' by default, so build will always happen.
        is_publish_allowed = is_publish_branch(ctx, hopic_git_info)
        volumes = cfg['volumes'].copy()
        global_timeout = None
        global_timeout_expire_time = None
        finally_cmds = {"sh-bound": [], "global": determine_global_finally_block(cmds)}

        for cmd in cmds:
            worktrees = {}
            foreach = None

            assert isinstance(cmd, Mapping)

            run_on_change = cmd.get('run-on-change', RunOnChange.default)
            if run_on_change == RunOnChange.always:
                pass
            elif run_on_change == RunOnChange.never:
                if hopic_git_info.has_change:
                    break
            elif run_on_change in (RunOnChange.only, RunOnChange.new_version_only):
                if not hopic_git_info.has_change:
                    log.debug("No change detected for %r", hopic_git_info)
                    break
                if not is_publish_allowed:
                    log.debug("Not allowed to publish for %r", hopic_git_info)
                    break
                if run_on_change == RunOnChange.new_version_only and not hopic_git_info.version_bumped:
                    log.debug("No version change detected for %r", hopic_git_info)
                    break
            try:
                desc = cmd['description']
            except (KeyError, TypeError):
                pass
            else:
                log.info('Performing: %s', click.style(desc, fg='cyan'))

            monotonic_now = time.monotonic()
            if "sh" not in cmd and "timeout" in cmd:
                assert global_timeout is None, "there should only be a single global per-variant timeout and config_reader should have enforced that"
                global_timeout = cmd["timeout"]
                global_timeout_expire_time = monotonic_now + global_timeout
                log.debug("restricting all commands combined to a maximum time of %f seconds", cmd["timeout"])

            timeout = None
            if "sh" in cmd:
                if "timeout" in cmd:
                    timeout = cmd["timeout"]
                global_timeout_remainder = None
                if global_timeout_expire_time is not None:
                    global_timeout_remainder = global_timeout_expire_time - monotonic_now
                    if global_timeout_remainder <= 0:
                        raise StepTimeoutExpiredError(global_timeout, cmd=" ".join(cmd["sh"]), before=True)

                if global_timeout_remainder is not None and timeout is None:
                    timeout = global_timeout_remainder
                    log.debug("restricting current command to a maximum of %f seconds remaining from global timeout", timeout)
                elif global_timeout_remainder is not None and timeout > global_timeout_remainder:
                    log.debug(
                        "restricting current command to a maximum of %f seconds remaining from global timeout (clamped from %f)",
                        global_timeout_remainder,
                        timeout,
                    )
                elif timeout is not None:
                    log.debug("restricting current command to a maximum of %f seconds", timeout)

                if "finally" in cmd:
                    finally_cmds["sh-bound"].extend(cmd["finally"])

            try:
                cmd_volumes_from = cmd['volumes-from']
            except (KeyError, TypeError):
                pass
            else:
                if image:
                    for volume in cmd_volumes_from:
                        volumes_from.add(volume['image'])
                else:
                    log.warning('`volumes-from` has no effect if no Docker image is configured')

            for artifact_key in (
                        'archive',
                        'fingerprint',
                    ):
                try:
                    opt = cmd[artifact_key]
                except (KeyError, TypeError):
                    pass
                else:
                    if artifact_key == "junit":
                        new_artifacts = opt["test-results"]
                    else:
                        new_artifacts = (artifact["pattern"].replace("(*)", "*") for artifact in opt["artifacts"])
                    if opt["allow-missing"]:
                        optional_artifacts.extend(new_artifacts)
                    else:
                        mandatory_artifacts.extend(new_artifacts)
            try:
                junit = cmd["junit"]
            except (KeyError, TypeError):
                pass
            else:
                if not junit["allow-missing"]:
                    mandatory_junit.extend(junit["test-results"])

            if not ctx.obj.dry_run:
                try:
                    worktrees = cmd['worktrees']
                except KeyError:
                    pass
                else:
                    # Force clean builds when we don't know how to discover changed files
                    for subdir, worktree in worktrees.items():
                        if worktree["changed-files"] is None:
                            with git.Repo(ctx.obj.workspace / subdir) as repo:
                                clean_output = repo.git.clean(subdir, x=True, d=True, force=True)
                                if clean_output:
                                    log.info('%s', clean_output)

            changed_files: typing.Sequence[PathLike] = cmd.get("changed-files", ())

            try:
                foreach = cmd['foreach']
            except KeyError:
                pass

            try:
                scoped_volumes = expand_docker_volume_spec(ctx.obj.config_dir,
                                                           ctx.obj.volume_vars, cmd['volumes'],
                                                           add_defaults=False)
                volumes.update(scoped_volumes)
            except KeyError:
                pass

            try:
                image = cmd['image']
            except KeyError:
                pass

            cmd_extra_docker_run_args = cmd.get('extra-docker-args', '')
            if cmd_extra_docker_run_args:
                if not image:
                    log.warning('`extra-docker-args` has no effect if no Docker image is configured')
                else:
                    for arg in cmd_extra_docker_run_args:
                        value = cmd_extra_docker_run_args[arg]
                        if isinstance(value, bool):
                            if not value:
                                log.warning('A "False" value for an `extra-docker-args` argument has no meaning and will be ignored')
                            else:
                                extra_docker_run_args.append(f'--{arg}')
                        else:
                            if (isinstance(value, Sequence) and not isinstance(value, str)):
                                extra_docker_run_args.extend((f'--{arg}={v}' for v in value))
                            else:
                                extra_docker_run_args.append(f'--{arg}={value}')

            try:
                docker_in_docker = cmd['docker-in-docker']
            except KeyError:
                pass

            try:
                with_credentials = cmd['with-credentials']
            except (KeyError, TypeError):
                pass
            else:
                for creds in with_credentials:
                    if 'project-name' in cfg and creds['type'] == CredentialType.username_password and not (
                            creds['username-variable'] in ctx.obj.volume_vars
                            and creds['password-variable'] in ctx.obj.volume_vars):
                        if ctx.obj.dry_run:
                            # Use dummy values for dry runs instead of (potentially) asking the user
                            kcred = (f'user_{id}', f'pass_{id}')
                        else:
                            kcred = credentials.get_credential_by_id(cfg['project-name'], creds['id'])

                        if kcred is not None:
                            username, password = kcred
                            if creds['encoding'] == CredentialEncoding.url:
                                username = urllib.parse.quote(username, safe='')
                                password = urllib.parse.quote(password, safe='')
                            volume_vars.update({
                                creds['username-variable']: username,
                                creds['password-variable']: password,
                            })

                    cred_vars = {name for key, name in creds.items() if key.endswith('-variable')}
                    for cred_var in cred_vars:
                        if cred_var in volume_vars:
                            if not isinstance(volume_vars[cred_var], MissingCredentialVarError):
                                variant_credentials[cred_var] = volume_vars[cred_var]
                        else:
                            volume_vars[cred_var] = MissingCredentialVarError(creds['id'], cred_var)

            try:
                cmd_env = cmd['environment']
                cmd = cmd['sh']
            except KeyError:
                continue

            volume_vars["WORKSPACE"] = "/code" if image is not None else str(ctx.obj.code_dir)
            volume_vars["CFGDIR"] = "/cfg" if image is not None else str(ctx.obj.config_dir)

            env = (dict(
                HOME            = '/home/sandbox',              # noqa: E251 "unexpected spaces around '='"
                _JAVA_OPTIONS   = '-Duser.home=/home/sandbox',  # noqa: E251 "unexpected spaces around '='"
            ) if image is not None else {})

            # Force or suppress colors according to http://bixense.com/clicolors/ convention.
            color: Optional[bool] = ctx.color
            if color is None:
                # When Hopic is only using auto detection then only pass on the existing values of these variables.
                if "CLICOLOR" in os.environ:
                    env["CLICOLOR"] = os.environ["CLICOLOR"]
                if "CLICOLOR_FORCE" in os.environ:
                    env["CLICOLOR_FORCE"] = os.environ["CLICOLOR_FORCE"]
            elif color:
                # CLICOLOR_FORCE overrules CLICOLOR so we can ignore the latter in this case
                env["CLICOLOR_FORCE"] = "1"
            else:
                assert not ctx.color
                env["CLICOLOR_FORCE"] = "0"
                env["CLICOLOR"] = "0"

            for varname in cfg['pass-through-environment-vars']:
                if varname in os.environ:
                    env.setdefault(varname, os.environ[varname])

            for varname in (
                        'SOURCE_DATE_EPOCH',
                        'VERSION',
                        'PURE_VERSION',
                        'DEBVERSION',
                        'PUBLISH_VERSION'
                    ):
                if varname in ctx.obj.volume_vars and not isinstance(ctx.obj.volume_vars[varname], BaseException):
                    env[varname] = ctx.obj.volume_vars[varname]

            env.update(variant_credentials)
            foreach_items = (None,)
            if foreach == 'SOURCE_COMMIT':
                foreach_items = hopic_git_info.source_commits
            elif foreach == 'AUTOSQUASHED_COMMIT':
                foreach_items = hopic_git_info.autosquashed_commits

            for foreach_item in foreach_items:
                cfg_vars = volume_vars.copy()
                if foreach in (
                            'SOURCE_COMMIT',
                            'AUTOSQUASHED_COMMIT',
                        ):
                    cfg_vars[foreach] = str(foreach_item)
                if git_commit_time is not None:
                    try:
                        # Be compatible with reproducible-builds.org
                        now = datetime.utcfromtimestamp(float(os.environ["SOURCE_DATE_EPOCH"])).replace(tzinfo=tzutc())
                    except (KeyError, TypeError):
                        now = datetime.utcnow().replace(tzinfo=tzutc())
                    duration = now - git_commit_time
                    cfg_vars["BUILD_DURATION"] = f"{duration.total_seconds():.6f}"

                final_env = env.copy()
                for k, v in cmd_env.items():
                    if v is None and k in final_env:
                        del final_env[k]
                    else:
                        final_env[k] = expand_vars(cfg_vars, v)
                final_cmd = [expand_vars(cfg_vars, arg) for arg in cmd]

                if final_cmd == [":"]:
                    # NOP: skip. This command, on *nix, would always do nothing and return with exit code 0. So abuse it for a NOP.
                    continue

                # Handle execution inside docker
                cidfile = None
                try:
                    if image is not None:
                        uid, gid = os.getuid(), os.getgid()
                        fd, cidfile = tempfile.mkstemp(prefix='hopic-docker-run-cid-', suffix='.txt')
                        os.close(fd)
                        # Docker wants this file to not exist (yet) when starting a container
                        os.unlink(cidfile)
                        docker_run = [
                            "docker",
                            "run",
                            "--rm",
                            f"--cidfile={cidfile}",
                            "--net=host",
                            "--cap-add=SYS_PTRACE",
                            f"--tmpfs={final_env['HOME']}:exec,uid={uid},gid={gid}",
                            f"--user={uid}:{gid}",
                            f"--workdir={expand_vars(volume_vars, cwd)}",
                            *(f"--env={k}={v}" for k, v in final_env.items()),
                        ]

                        if all(hasattr(fd, 'isatty') and fd.isatty() for fd in [sys.stderr, sys.stdout, sys.stdin]):
                            docker_run += ['--tty']

                        if docker_in_docker:
                            try:
                                sock = '/var/run/docker.sock'
                                st = os.stat(sock)
                            except OSError as e:
                                log.error("Docker in Docker access requested but cannot access Docker socket: %s", e)
                            else:
                                if stat.S_ISSOCK(st.st_mode):
                                    docker_run += [f"--volume={sock}:{sock}"]
                                    # Give group access to the socket if it's group accessible but not world accessible
                                    if st.st_mode & 0o0060 == 0o0060 and st.st_mode & 0o0006 != 0o0006:
                                        docker_run += [f"--group-add={st.st_gid}"]

                        for volume in volumes.values():
                            docker_run += ['--volume={}'.format(volume_spec_to_docker_param(volume))]

                        for volume_from in volumes_from:
                            docker_run += ['--volumes-from=' + volume_from]

                        docker_run += extra_docker_run_args
                        docker_run.append(str(image))
                        final_cmd = docker_run + final_cmd
                    new_env = os.environ.copy()
                    if image is None:
                        new_env.update(final_env)
                        for k, v in cmd_env.items():
                            if v is None and k in new_env:
                                del new_env[k]

                    def signal_handler(signum, frame):
                        log.warning('Received fatal signal %d', signum)
                        raise FatalSignal(signum)

                    old_handlers = dict((num, signal.signal(num, signal_handler)) for num in (signal.SIGINT, signal.SIGTERM))
                    try:
                        run_method: Callable[..., Any] = subprocess.check_call
                        run_args = {}
                        if exec_stdout is subprocess.PIPE:
                            run_method = subprocess.check_output
                        elif exec_stdout is not None:
                            run_args["stdout"] = exec_stdout

                        yield echo_cmd(
                            run_method,
                            final_cmd,
                            env=new_env,
                            cwd=expand_vars(ctx.obj.volume_vars, cwd),
                            obfuscate=variant_credentials,
                            timeout=timeout,
                            **run_args,
                        )
                    except subprocess.CalledProcessError as e:
                        handle_finally(finally_cmds["sh-bound"], finally_cmds["global"], variant, hopic_git_info, config_based_volume_vars)
                        log.error("Command fatally terminated with exit code %d", e.returncode)
                        ctx.exit(e.returncode)
                    except (FatalSignal, subprocess.TimeoutExpired) as exc:
                        handle_finally(finally_cmds["sh-bound"], finally_cmds["global"], variant, hopic_git_info, config_based_volume_vars)
                        if cidfile and os.path.isfile(cidfile):
                            # If we're being signalled to shut down ensure the spawned docker container also gets cleaned up.
                            with open(cidfile) as f:
                                cid = f.read()
                            try:
                                # Will also remove the container due to the '--rm' it was started with.
                                echo_cmd(subprocess.check_call, ('docker', 'stop', cid))
                            except subprocess.CalledProcessError as e:
                                log.error(
                                        'Could not stop Docker container (maybe it was stopped already?), command failed with exit code %d',
                                        e.returncode)
                            except FatalSignal:
                                # Ignore INT and TERM while sending KILL to the Docker container
                                signal.signal(signal.SIGINT, signal.SIG_IGN)
                                signal.signal(signal.SIGTERM, signal.SIG_IGN)
                                log.warning('Interrupted while stopping Docker container; killing..')
                                echo_cmd(subprocess.check_call, ('docker', 'kill', cid))
                        if isinstance(exc, FatalSignal):
                            ctx.exit(128 + exc.signal)
                        else:
                            assert isinstance(exc, subprocess.TimeoutExpired)
                            raise StepTimeoutExpiredError(timeout, cmd=" ".join(cmd))
                    except Exception as e:
                        handle_finally(finally_cmds["sh-bound"], finally_cmds["global"], variant, hopic_git_info, config_based_volume_vars)
                        raise e
                    for num, old_handler in old_handlers.items():
                        signal.signal(num, old_handler)
                finally:
                    if cidfile:
                        try:
                            os.unlink(cidfile)
                        except FileNotFoundError:
                            pass

            with git.Repo(ctx.obj.workspace) as repo:
                source_commit = repo.head.commit
                if changed_files:
                    repo.index.add(expand_vars(volume_vars, f) for f in changed_files)

            for subdir, worktree in worktrees.items():
                with git.Repo(ctx.obj.workspace / subdir) as repo:
                    worktree_commits.setdefault(subdir, [
                        str(repo.head.commit),
                        str(repo.head.commit),
                    ])

                    changed_files = worktree["changed-files"]
                    if changed_files is not None:
                        repo.index.add(expand_vars(volume_vars, f) for f in changed_files)
                    else:
                        # 'git add --all' equivalent (excluding the code_dir)
                        add_files = set(repo.untracked_files)
                        remove_files = set()
                        for diff in repo.index.diff(None):
                            if not diff.deleted_file:
                                add_files.add(diff.b_path)
                            remove_files.add(diff.a_path)
                        remove_files -= add_files
                        if remove_files:
                            repo.index.remove(remove_files)
                        if add_files:
                            repo.index.add(add_files)

                    commit_message = expand_vars(volume_vars, worktree['commit-message'])
                    if not commit_message.endswith(u'\n'):
                        commit_message += u'\n'
                    # fmt: off
                    submit_commit = repo.index.commit(
                        message     = commit_message,                                 # noqa: E251 "unexpected spaces around '='"
                        author      = source_commit.author,                           # noqa: E251 "unexpected spaces around '='"
                        committer   = source_commit.committer,                        # noqa: E251 "unexpected spaces around '='"
                        author_date = to_git_time(source_commit.authored_datetime),   # noqa: E251 "unexpected spaces around '='"
                        commit_date = to_git_time(source_commit.committed_datetime),  # noqa: E251 "unexpected spaces around '='"
                    )
                    # fmt: on
                    restore_mtime_from_git(repo)
                    worktree_commits[subdir][1] = str(submit_commit)
                    log.info('%s', repo.git.show(submit_commit, format='fuller', stat=True))

        handle_finally(finally_cmds["sh-bound"], finally_cmds["global"], variant, hopic_git_info, config_based_volume_vars)

        if worktree_commits:
            with git.Repo(ctx.obj.workspace) as repo, repo.config_writer() as git_cfg:
                submit_commit = repo.head.commit
                section = f"hopic.{submit_commit}"
                if git_cfg.has_option(section, 'refspecs'):
                    refspecs = list(shlex.split(git_cfg.get_value(section, 'refspecs')))
                else:
                    refspecs = []

                bundle_commits = []
                for subdir, (base_commit, submit_commit) in worktree_commits.items():
                    worktree_ref = ctx.obj.config['scm']['git']['worktrees'][subdir]
                    if worktree_ref in repo.heads:
                        repo.heads[worktree_ref].set_commit(submit_commit, logmsg='Prepare for git-bundle')
                    else:
                        repo.create_head(worktree_ref, submit_commit)
                    bundle_commits.append(f"{base_commit}..{worktree_ref}")
                    refspecs.append(f"{submit_commit}:{worktree_ref}")
                repo.git.bundle('create', ctx.obj.workspace / 'worktree-transfer.bundle', *bundle_commits)

                git_cfg.set_value(section, 'refspecs', ' '.join(shlex.quote(refspec) for refspec in refspecs))

        # Post-processing to make these artifacts as reproducible as possible
        for artifact_pattern in optional_artifacts:
            for artifact in ctx.obj.code_dir.glob(artifact_pattern):
                binary_normalize.normalize(artifact, source_date_epoch=ctx.obj.source_date_epoch)

        pattern_matched = False
        mandatory_artifacts = [expand_vars(volume_vars, exp) for exp in mandatory_artifacts]
        for pattern in mandatory_artifacts:
            for artifact in ctx.obj.code_dir.glob(pattern):
                pattern_matched = True
                binary_normalize.normalize(artifact, source_date_epoch=ctx.obj.source_date_epoch)
        if mandatory_artifacts and not pattern_matched:
            raise MissingFileError(f"none of these mandatory artifact patterns matched a file: {mandatory_artifacts}")

        pattern_matched = False
        mandatory_junit = [expand_vars(volume_vars, exp) for exp in mandatory_junit]
        for pattern in mandatory_junit:
            for _ in ctx.obj.code_dir.glob(pattern):
                pattern_matched = True
                break
        if mandatory_junit and not pattern_matched:
            raise MissingFileError(f"none of these mandatory junit patterns matched a file: {mandatory_junit}")


@click.command()
@click.option('--phase'  , '-p', metavar='<phase>'  , multiple=True, help='''Build phase to execute''', autocompletion=autocomplete.phase_from_config)
@click.option('--variant', '-v', metavar='<variant>', multiple=True, help='''Configuration variant to build''', autocompletion=autocomplete.variant_from_config)
@click.option('--dry-run', '-n', is_flag=True, default=False, help='''Print commands from the configured phases and variants, but do not execute them''')
@click.pass_context
def build(ctx, phase, variant, dry_run):
    """
    Build for the specified commit.

    This defaults to building all variants for all phases.
    It's possible to limit building to either all variants for a single phase, all phases for a single variant or a
    single variant for a single phase.
    """
    # Ensure any required extensions are available
    initialize_global_variables_from_config(extensions.install_extensions.callback())

    ctx.obj.dry_run = dry_run
    if dry_run:
        log.info('[dry-run] would execute:')

    hopic_git_info = HopicGitInfo.from_repo(ctx.obj.workspace)

    unknown_phases = [phasename for phasename in phase if phasename not in ctx.obj.config['phases']]
    if unknown_phases:
        raise UnknownPhaseError(phase=unknown_phases)

    phases_obj = ctx.obj.config["phases"]
    next_phases = {}
    for phasename in phases_obj.keys():
        for next_phase in next_phases:
            next_phases[next_phase].append(phasename)
        next_phases[phasename] = []

    for phasename, curphase in phases_obj.items():
        if phase and phasename not in phase:
            continue

        for var in variant:
            if var not in curphase:
                log.warning(f"phase '{phasename}' does not contain variant '{var}'")

        for curvariant, cmds in curphase.items():
            if variant and curvariant not in variant:
                continue

            try:
                (*_,) = build_variant(
                    variant=curvariant,
                    cmds=cmds,
                    hopic_git_info=hopic_git_info,
                    config_based_volume_vars={"HOPIC_PHASE": phasename, "HOPIC_VARIANT": curvariant},
                )
            except Exception:
                for next_phase in next_phases[phasename]:
                    if curvariant not in phases_obj[next_phase] or (
                        not any(
                            "wait-on-full-previous-phase" in elem and elem["wait-on-full-previous-phase"] is False
                            for elem in phases_obj[next_phase][curvariant]
                        )
                    ):
                        break
                    # Chained variants

                    handle_finally(
                        [],
                        determine_global_finally_block(phases_obj[next_phase][curvariant]),
                        curvariant,
                        hopic_git_info,
                        {"HOPIC_PHASE": next_phase, "HOPIC_VARIANT": curvariant},
                    )
                raise


def handle_finally(cmd_finally_cmds, global_finally_cmds, variant, hopic_git_info, config_based_volume_vars):
    try:
        if cmd_finally_cmds:
            (*_,) = build_variant(f"{variant}-finally", cmd_finally_cmds, hopic_git_info, config_based_volume_vars=config_based_volume_vars)
    except Exception:
        log.warning("error in %s-finally:", variant, exc_info=True)
        raise
    finally:
        if global_finally_cmds:
            (*_,) = build_variant(f"{variant}-finally", global_finally_cmds, hopic_git_info, config_based_volume_vars=config_based_volume_vars)


def determine_global_finally_block(cmds):
    global_finally_blocks = [cmd["finally"] for cmd in cmds if "finally" in cmd and "sh" not in cmd]
    return [cmd for finally_block in global_finally_blocks for cmd in finally_block]



# Copyright (c) 2018 - 2019 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import logging
import functools
import locale
import os
import re
import shlex

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def no_exec(*args, **kwargs):
    return 0


# Caching is not just an optimization: manipulating the locale multiple times is risky on some platforms
@functools.lru_cache(maxsize=None)
def determine_locale_envvars():
    # Prefer the current LC_CTYPE (character encoding) locale when it's already UTF-8
    if re.match(".*utf[-_]?8$", os.environ.get("LC_CTYPE", ""), re.IGNORECASE):
        return {"LANG": "C", "LC_CTYPE": os.environ["LC_CTYPE"]}

    # FreeBSD's preferred way of setting the encoding is to specify only the codeset name in LC_CTYPE without any language (and MacOS inherits this).
    # This doesn't work on glibc based operating systems though (most Linux distros) so make it optional.
    try:
        locale.setlocale(locale.LC_CTYPE, "UTF-8")
    except locale.Error:
        pass
    else:
        return {"LANG": "C", "LC_CTYPE": "UTF-8"}

    # C.UTF-8 works on all Linux distros (glibc really)
    # Also LC_ALL has precedence over all other LC_* environment variables, so overrides the rest.
    # So we can use that only if we're not setting any other LC_* vars.
    return {"LC_ALL": "C.UTF-8"}


def echo_cmd(fun, cmd, *args, dry_run=False, obfuscate=None, **kwargs):
    command_list = []
    for word in cmd:
        if obfuscate is not None:
            for secret_name, secret in obfuscate.items():
                if secret and isinstance(secret, str):
                    word = word.replace(secret, f'${{{secret_name}}}')
        command_list.append(shlex.quote(word))
    log.info('%s%s', '' if dry_run else 'Executing: ',
             click.style(' '.join(command_list), fg='yellow'))

    # Set our locale for machine readability with UTF-8
    kwargs = kwargs.copy()
    try:
        env = kwargs['env'].copy()
    except KeyError:
        env = os.environ.copy()
    for key in list(env):
        if key.startswith("LC_") or key in ("LANG", "LANGUAGE"):
            del env[key]
    env.update(determine_locale_envvars())
    kwargs['env'] = env

    try:
        exec_fun = no_exec if dry_run else fun
        output = exec_fun(cmd, *args, **kwargs)

        return (output.decode('UTF-8') if isinstance(output, bytes) else output)
    except Exception as e:
        if hasattr(e, 'child_traceback'):
            log.exception('Child traceback: %s', e.child_traceback)
        raise


def echo_cmd_click(fun, cmd, *args, obfuscate=None, **kwargs):
    ctx = click.get_current_context()
    return echo_cmd(fun, cmd, *args, **kwargs, obfuscate=obfuscate, dry_run=getattr(ctx.obj, 'dry_run', False))



# Copyright (c) 2018 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from . import (
        autocomplete,
        build,
        extensions,
        utils,
    )
from .utils import (
        determine_config_file_name,
        get_package_version,
        is_publish_branch,
    )
from commisery.commit import CommitMessage, parse_commit_message
from ..build import (
    HopicGitInfo,
)
from ..config_reader import (
        JSONEncoder,
        expand_vars,
        read as read_config,
    )
from ..execution import echo_cmd_click as echo_cmd
from ..git_time import (
    GitVersion,
    determine_git_version,
    determine_version,
    restore_mtime_from_git,
    to_git_time,
)
from .global_obj import initialize_global_variables_from_config
from ..versioning import (
    Version,
    hotfix_id,
    replace_version,
)
from collections import OrderedDict
from collections.abc import (
        Mapping,
        MutableMapping,
        MutableSequence,
        Set,
    )
from configparser import (
        NoOptionError,
        NoSectionError,
    )
from datetime import datetime
from dateutil.parser import parse as date_parse
from dateutil.tz import (tzoffset, tzlocal)
import git
import gitdb
from io import (
        BytesIO,
        StringIO,
    )
import json
import logging
import os
from pathlib import Path
import platform
import re
import shlex
import shutil
import subprocess
import sys
import typing
from typing import (
    AbstractSet,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)
from textwrap import dedent
from yaml.error import YAMLError

from .main import main
from ..errors import (
    CommitAncestorMismatchError,
    GitNotesMismatchError,
    VersionBumpMismatchError,
    VersioningError,
)
from ..types import PathLike


PACKAGE : str = __package__.split('.')[0]

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


for submodule in (
    build,
    extensions,
):
    for subcmd in dir(submodule):
        if subcmd.startswith('_'):
            continue
        cmd = getattr(submodule, subcmd)
        if not isinstance(cmd, click.Command):
            continue
        main.add_command(cmd)


class DateTime(click.ParamType):
    name = 'date'
    stamp_re = re.compile(r'^@(?P<utcstamp>\d+(?:\.\d+)?)(?:\s+(?P<tzdir>[-+])(?P<tzhour>\d{1,2}):?(?P<tzmin>\d{2}))?$')

    def convert(self, value, param, ctx):
        if value is None or isinstance(value, datetime):
            return value

        try:
            stamp = self.stamp_re.match(value)
            if stamp:
                def int_or_none(i):
                    if i is None:
                        return None
                    return int(i)

                tzdir  = (-1 if stamp.group('tzdir') == '-' else 1)
                tzhour = int_or_none(stamp.group('tzhour'))
                tzmin  = int_or_none(stamp.group('tzmin'))

                if tzhour is not None:
                    tz = tzoffset(None, tzdir * (tzhour * 3600 + tzmin * 60))
                else:
                    tz = tzlocal()
                return datetime.fromtimestamp(float(stamp.group('utcstamp')), tz)

            dt = date_parse(value)
            if dt.tzinfo is None:
                dt = dt.replace(tzinfo=tzlocal())
            return dt
        except ValueError as e:
            self.fail('Could not parse datetime string "{value}": {e}'.format(value=value, e=' '.join(e.args)), param, ctx)


@main.command()
@click.pass_context
def may_publish(ctx):
    """
    Check if the target branch name is allowed to be published, according to publish-from-branch in the config file.
    """

    ctx.exit(0 if is_publish_branch(ctx) else 1)


def checkout_tree(
    tree: PathLike,
    remote: Optional[str],
    ref: str,
    *,
    commit: Union[str, git.Commit, None] = None,
    clean: bool = False,
    remote_name: str = 'origin',
    tags: bool = True,
    allow_submodule_checkout_failure: bool = False,
    clean_config: Union[List, Tuple] = (),
):
    fresh = clean
    try:
        repo = git.Repo(tree)
        # Cleanup potential existing submodules to avoid conflicts in PR's where submodules are added
        # Cannot use config file here to determine if feature is enabled since config is not parsed during checkout-source-tree
        repo.git.submodule(["deinit", "--all", "--force"])
        modules_dir = "%s/modules" % repo.git_dir
        if os.path.isdir(modules_dir):
            # Hacky way to restore git repo to clean state
            shutil.rmtree(modules_dir)
    except (git.InvalidGitRepositoryError, git.NoSuchPathError):
        if clean and os.path.exists(tree):
            # Wipe the directory to allow 'git clone' to succeed.
            # It would fail if it isn't empty.

            # We're deleting only the content of the directory, because deleting 'tree' when it's the current working
            # directory of processes would cause getcwd(3) to fail.
            for name in os.listdir(tree):
                path = os.path.join(tree, name)
                if os.path.isdir(path):
                    shutil.rmtree(tree)
                else:
                    os.remove(path)

        assert remote is not None
        repo = git.Repo.clone_from(remote, tree)
        fresh = True

    with repo:
        with repo.config_writer() as cfg:
            cfg.remove_section('hopic.code')
            cfg.set_value('color', 'ui', 'always')
            cfg.set_value('hopic.code', 'cfg-clean', str(clean))
            cfg.set_value("hopic.code", "fresh", str(fresh))

        if remote is not None:
            clean_tags = tags and repo.tags
            if clean_tags:
                repo.delete_tag(*clean_tags)

            try:
                # Delete, instead of update, existing remotes.
                # This is because of https://github.com/gitpython-developers/GitPython/issues/719
                repo.delete_remote(remote_name)
            except git.GitCommandError:
                pass
            origin = repo.create_remote(remote_name, remote)

            fetch_info, *_ = origin.fetch(ref, tags=tags)

            if commit is not None and not repo.is_ancestor(commit, fetch_info.commit):
                raise CommitAncestorMismatchError(commit, fetch_info.commit, ref)

            commit = repo.commit(commit) if commit else fetch_info.commit

            # Ensure we have the exact same view of all Hopic notes as are present upstream
            origin.fetch("+refs/notes/hopic/*:refs/notes/hopic/*", prune=True)
        else:
            assert commit is not None
            if isinstance(commit, str):
                commit = repo.commit(commit)

        repo.head.reference = commit
        repo.head.reset(index=True, working_tree=True)
        # Remove potentially moved submodules
        repo.git.submodule(["deinit", "--all", "--force"])

        try:
            update_submodules(repo, clean)
        except git.GitCommandError as e:
            log.error(dedent("""\
                    Failed to checkout submodule for ref '%s'
                    error:
                    %s"""), ref, e)
            if not allow_submodule_checkout_failure:
                raise

        if clean:
            clean_repo(repo, clean_config)
        if fresh:
            # Only restore mtimes when doing a clean build or working on a fresh clone. This prevents problems with timestamp-based build sytems.
            # I.e. make and ninja and probably half the world.
            restore_mtime_from_git(repo)

        with repo.config_writer() as cfg:
            section = f"hopic.{commit}"
            cfg.set_value(section, 'ref', ref)
            if remote is not None:
                cfg.set_value(section, "remote", remote)

    return commit


def update_submodules(repo, clean):
    for submodule in repo.submodules:
        log.info("Updating submodule: %s and clean = %s" % (submodule, clean))
        repo.git.submodule(["sync", "--recursive"])
        # Cannot use submodule.update call here since this call doesn't use git submodules call
        # It tries to emulate the behaviour with a git clone call, but this doesn't work with relative submodule URL's
        # See https://github.com/gitpython-developers/GitPython/issues/944
        repo.git.submodule(["update", "--init", "--recursive"])

        with git.Repo(os.path.join(repo.working_dir, submodule.path)) as sub_repo:
            update_submodules(sub_repo, clean)
            if clean:
                clean_repo(sub_repo)


def clean_repo(repo, clean_config=[]):
    def substitute_home(arg):
        volume_vars = {'HOME': os.path.expanduser('~')}
        return expand_vars(volume_vars, os.path.expanduser(arg))
    for cmd in clean_config:
        cmd = [substitute_home(arg) for arg in shlex.split(cmd)]
        try:
            echo_cmd(subprocess.check_call, cmd, cwd=repo.working_dir)
        except subprocess.CalledProcessError as e:
            log.error("Command fatally terminated with exit code %d", e.returncode)
            sys.exit(e.returncode)

    clean_output = repo.git.clean(x=True, d=True, force=(True, True))
    if clean_output:
        log.info('%s', clean_output)


def install_extensions_and_parse_config(constraints: Optional[str] = None):
    initialize_global_variables_from_config(extensions.install_extensions.callback(constraints=constraints))


_code_dir_re = re.compile(r"^code(?:-\d+)$")


def find_code_dir(workspace: PathLike) -> Path:
    code_dirs = sorted(Path(dir) for dir in os.listdir(workspace) if _code_dir_re.match(dir))
    for dir in code_dirs:
        try:
            with git.Repo(workspace / dir):
                pass
        except (git.InvalidGitRepositoryError, git.NoSuchPathError):
            pass
        else:
            return dir
    else:
        seq = 0
        while True:
            dir = Path("code" if seq == 0 else f"code-{seq:03}")
            seq += 1
            if dir not in code_dirs:
                return dir


def checkout_worktrees(workspace: Path, worktrees: Dict[PathLike, str]):
    if not worktrees:
        return

    with git.Repo(workspace) as repo:
        fetch_result = repo.remotes.origin.fetch([ref for subdir, ref in worktrees.items()])

        worktree_refs = {Path(subdir): fetchinfo.ref for (subdir, refname), fetchinfo in zip(worktrees.items(), fetch_result)}
        log.debug("Worktree config: %s", worktree_refs)

        for subdir, ref in worktree_refs.items():
            try:
                os.remove(workspace / subdir / ".git")
            except (OSError, IOError):
                pass
            clean_output = repo.git.clean("-xd", subdir, force=True)
            if clean_output:
                log.info("%s", clean_output)

        repo.git.worktree("prune")

        for subdir, ref in worktree_refs.items():
            repo.git.worktree("add", subdir, ref.commit)


def store_commit_meta(repo: git.Repo, commit_meta: Dict[str, Any], *, commit: git.Commit, old_commit: Optional[git.Commit] = None) -> None:
    with repo.config_writer() as cfg:
        if old_commit is not None:
            cfg.remove_section(f"hopic.{old_commit}")
        section = f"hopic.{commit}"
        for key, val in commit_meta.items():
            if val is None:
                continue
            if isinstance(val, bool):
                cfg.set_value(section, key, "true" if val else "false")
            elif isinstance(val, git.Object):
                cfg.set_value(section, key, str(val))
            elif isinstance(val, list):
                cfg.set_value(section, "refspecs", " ".join(shlex.quote(refspec) for refspec in val))
            else:
                cfg.set_value(section, key, val)


@main.command()
@click.option('--target-remote'     , metavar='<url>')
@click.option('--target-ref'        , metavar='<ref>')
@click.option('--target-commit'     , metavar='<commit>')
@click.option('--clean/--no-clean'  , default=False, help='''Clean workspace of non-tracked files''')
@click.option('--ignore-initial-submodule-checkout-failure/--no-ignore-initial-submodule-checkout-failure',
              default=False, help='''Ignore git submodule errors during initial checkout''')
@click.pass_context
def checkout_source_tree(
    ctx,
    target_remote,
    target_ref,
    target_commit,
    clean,
    ignore_initial_submodule_checkout_failure,
):
    """
    Checks out a source tree of the specified remote's ref to the workspace.
    """

    workspace = ctx.obj.workspace
    # Check out specified repository
    click.echo(
        checkout_tree(
            workspace,
            target_remote,
            target_ref,
            commit=target_commit,
            clean=clean,
            allow_submodule_checkout_failure=ignore_initial_submodule_checkout_failure,
        )
    )

    try:
        ctx.obj.config = read_config(determine_config_file_name(ctx), ctx.obj.volume_vars)
        if clean:
            with git.Repo(workspace) as repo:
                clean_repo(repo, ctx.obj.config['clean'])
        git_cfg = ctx.obj.config['scm']['git']
    except (click.BadParameter, KeyError, TypeError, OSError, IOError, YAMLError):
        return

    checkout_worktrees(workspace, git_cfg["worktrees"])

    if 'remote' not in git_cfg and 'ref' not in git_cfg:
        return

    code_dir = find_code_dir(workspace)

    # Check out configured repository and mark it as the code directory of this one
    ctx.obj.code_dir = workspace / code_dir
    with git.Repo(workspace) as repo, repo.config_writer() as cfg:
        cfg.remove_section('hopic.code')
        cfg.set_value('hopic.code', 'dir', str(code_dir))
        cfg.set_value('hopic.code', 'cfg-remote', target_remote)
        cfg.set_value('hopic.code', 'cfg-ref', target_ref)
        cfg.set_value('hopic.code', 'cfg-clean', str(clean))

    checkout_tree(
        ctx.obj.code_dir,
        git_cfg.get("remote", target_remote),
        git_cfg.get("ref", target_ref),
        clean=clean,
        clean_config=ctx.obj.config["clean"],
    )


@main.group()
# fmt: off
# git
@click.option('--author-name'               , metavar='<name>'                 , help='''Name of change-request's author''')
@click.option('--author-email'              , metavar='<email>'                , help='''E-mail address of change-request's author''')
@click.option('--author-date'               , metavar='<date>', type=DateTime(), help='''Time of last update to the change-request''')
@click.option('--commit-date'               , metavar='<date>', type=DateTime(), help='''Time of starting to build this change-request''')
@click.option("--bundle"                    , metavar="<file>", type=click.Path(file_okay=True, dir_okay=False, writable=True))
@click.option("--constraints"               , metavar="<file>", type=click.Path(exists=True, dir_okay=False, readable=True), help="""Apply the provided constraints file to pip operations""")
# fmt: on
def prepare_source_tree(*args, **kwargs):
    """
    Prepares the source tree for building a change performed by a subcommand.
    """

    pass


@prepare_source_tree.resultcallback()
@click.pass_context
def process_prepare_source_tree(
    ctx,
    change_applicator,
    author_name,
    author_email,
    author_date,
    commit_date,
    bundle: Optional[PathLike],
    constraints: Optional[str] = None,
):
    ctx.obj.pip_constraints = constraints

    with git.Repo(ctx.obj.workspace) as repo:
        if author_name is None or author_email is None:
            # This relies on /etc/passwd entries as a fallback, which might contain the info we need
            # for the current UID. Hence the conditional.
            author = git.Actor.author(repo.config_reader())
            if author_name is not None:
                author.name = author_name
            if author_email is not None:
                author.email = author_email
        else:
            author = git.Actor(author_name, author_email)

        try:
            committer = git.Actor.committer(repo.config_reader())
        except KeyError:
            committer = git.Actor(None, None)
        if not committer.name:
            committer.name = author.name
        if not committer.email:
            committer.email = author.email

        target_commit = repo.head.commit

        with repo.config_reader() as cfg:
            section = f"hopic.{target_commit}"
            target_ref    = cfg.get(section, 'ref', fallback=None)
            target_remote = cfg.get(section, 'remote', fallback=None)
            code_clean    = cfg.getboolean('hopic.code', 'cfg-clean', fallback=False)
            code_fresh = cfg.getboolean("hopic.code", "fresh", fallback=code_clean)

        repo.git.submodule(["deinit", "--all", "--force"])  # Remove submodules in case it is changed in change_applicator
        commit_params = change_applicator(repo, author=author, committer=committer)
        if not commit_params:
            return
        source_commit = commit_params.pop('source_commit', None)
        base_commit = commit_params.pop('base_commit', None)
        bump_message = commit_params.pop('bump_message', None)

        # Re-read config when it was not read already to ensure any changes introduced by 'change_applicator' are taken into account
        if not commit_params.pop('config_parsed', False):
            install_extensions_and_parse_config(ctx.obj.pip_constraints)

        # Ensure that, when we're dealing with a separated config and code repository, that the code repository is checked out again to the newer version
        if ctx.obj.code_dir != ctx.obj.workspace:
            with repo.config_reader() as cfg:
                try:
                    code_remote = ctx.obj.config['scm']['git']['remote']
                except (KeyError, TypeError):
                    code_remote = cfg.get_value('hopic.code', 'cfg-remote')
                try:
                    code_ref = ctx.obj.config["scm"]["git"]["ref"]
                except (KeyError, TypeError):
                    code_ref = cfg.get_value("hopic.code", "cfg-ref")

            checkout_tree(
                ctx.obj.code_dir,
                code_remote,
                code_ref,
                clean=code_clean,
                clean_config=ctx.obj.config["clean"],
            )

        version_info = ctx.obj.config['version']

        # Re-read version to ensure that the version policy in the reloaded configuration is used for it
        ctx.obj.version, _ = determine_version(version_info, ctx.obj.config_dir, ctx.obj.code_dir)

        # If the branch is not allowed to publish, skip version bump step
        is_publish_allowed = is_publish_branch(ctx)

        if 'file' in version_info:
            relative_version_file = os.path.relpath(os.path.join(os.path.relpath(ctx.obj.config_dir, repo.working_dir), version_info['file']))

        bump = version_info['bump'].copy()
        bump.update(commit_params.pop('bump-override', {}))

        commit_from, commit_to = (base_commit, target_commit) if base_commit else (target_commit, source_commit)
        source_commits = tuple(parse_commit_range(repo, commit_from, commit_to, bump))

        change_message = None
        if "message" in commit_params and bump["on-every-change"]:
            change_message = parse_commit_message(commit_params["message"], policy=bump["policy"], strict=bump.get("strict", False))

        hotfix = hotfix_id(version_info["hotfix-branch"], target_ref)

        def _is_valid_hotfix_base(version) -> bool:
            if not version.prerelease:
                # full release: valid point to start a hotfix from
                return True
            # Pre-release must be a valid hotfix prefix for the current hotfix ID
            return version.prerelease[:len(hotfix) + 1] == ("hotfix", *hotfix)

        if bump['policy'] == 'conventional-commits' and target_ref is not None:
            for commit in source_commits:
                if commit.has_breaking_change():
                    if bump['reject-breaking-changes-on'].match(target_ref):
                        raise VersioningError(
                                f"Breaking changes are not allowed on '{target_ref}', but commit '{commit.hexsha}' contains one:\n{commit.message}")
                    elif hotfix:
                        raise VersioningError(
                            f"Breaking changes are not allowed on hotfix branch '{target_ref}', but commit '{commit.hexsha}' contains one:\n{commit.message}")
                if commit.has_new_feature():
                    if bump['reject-new-features-on'].match(target_ref):
                        raise VersioningError(f"New features are not allowed on '{target_ref}', but commit '{commit.hexsha}' contains one:\n{commit.message}")
                    elif hotfix:
                        raise VersioningError(
                            f"New features are not allowed on hotfix branch '{target_ref}', but commit '{commit.hexsha}' contains one:\n{commit.message}")

        version_bumped = False
        if is_version_bump_enabled(bump, is_publish_from_branch_allowed=is_publish_allowed):
            cur_version = get_current_version(ctx)
            if hotfix:
                base_version = cur_version

                if "file" not in version_info:
                    with git.Repo(ctx.obj.code_dir) as code_repo:
                        gitversion = determine_git_version(code_repo)

                    params = {}
                    try:
                        params["format"] = version_info["format"]
                    except KeyError:
                        pass

                    # strip dirty state from version to ensure we're not complaining about that in the _is_valid_hotfix_base check below
                    commit_count = gitversion.commit_count
                    log.debug("checking for %r in %s..%s", version_info["hotfix-allowed-start-tags"], gitversion.tag_name, gitversion.commit_hash)
                    for commit in parse_commit_range(repo, gitversion.tag_name, gitversion.commit_hash, bump):
                        log.debug(
                            "checking whether %r of %r appears in %r",
                            getattr(commit, "type_tag", None),
                            commit.message,
                            version_info["hotfix-allowed-start-tags"],
                        )
                        if getattr(commit, "type_tag", None) in version_info["hotfix-allowed-start-tags"]:
                            commit_count -= 1
                    gitversion = GitVersion(tag_name=gitversion.tag_name, commit_hash=gitversion.commit_hash, commit_count=commit_count)
                    parse_version = gitversion.to_version(**params)
                    assert parse_version is not None
                    base_version = parse_version

                    # strip commit distance from version to ensure we're bumping the hotfix suffix instead of the commit distance suffix
                    gitversion = GitVersion(tag_name=gitversion.tag_name, commit_hash=gitversion.commit_hash)
                    parse_version = gitversion.to_version(**params)
                    assert parse_version is not None
                    cur_version = parse_version

                if not _is_valid_hotfix_base(base_version):
                    raise VersioningError(f"Creating hotfixes on anything but a full release is not supported. Currently on: {base_version}")

                release_part = str(cur_version.without_meta())
                if re.search(f"\\b{re.escape(release_part)}\\b", str(hotfix)):
                    raise VersioningError(f"Hotfix ID '{hotfix}' is not allowed to contain the base version '{release_part}'")

            if bump['policy'] == 'constant':
                params = {}
                if 'field' in bump:
                    params['bump'] = bump['field']
                if hotfix:
                    params["bump"] = "prerelease"
                    params["prerelease_seed"] = ("hotfix", *hotfix)
                    assert _is_valid_hotfix_base(cur_version), "implementation error: invalid hotfix bases should have been caught already"
                new_version = cur_version.next_version(**params)
            elif bump['policy'] in ('conventional-commits',):
                all_commits = source_commits
                if change_message is not None:
                    all_commits = (*source_commits, change_message)

                if log.isEnabledFor(logging.DEBUG):
                    log.debug("bumping based on conventional commits:")
                    for commit in all_commits:
                        breaking = ('breaking' if commit.has_breaking_change() else '')
                        feat = ('feat' if commit.has_new_feature() else '')
                        fix = ('fix' if commit.has_fix() else '')
                        try:
                            hash_prefix = click.style(commit.hexsha, fg='yellow') + ': '
                        except AttributeError:
                            hash_prefix = ''
                        log.debug("%s[%-8s][%-4s][%-3s]: %s", hash_prefix, breaking, feat, fix, commit.full_subject)
                new_version = cur_version.next_version_for_commits(all_commits)
                if hotfix and new_version != cur_version:
                    assert (new_version.major, new_version.minor) == (cur_version.major, cur_version.minor), (
                        "bumping anything other than 'patch' shouldn't happen for hotfix branches and should have been caught already"
                    )
                    assert _is_valid_hotfix_base(cur_version), "implementation error: invalid hotfix bases should have been caught already"
                    new_version = cur_version.next_prerelease(seed=("hotfix", *hotfix))
            else:
                raise NotImplementedError(f"unsupported version bumping policy {bump['policy']}")

            assert new_version >= cur_version, f"the new version {new_version} should be more recent than the old one {cur_version}"

            if new_version != cur_version:
                log.info("bumped version to: %s (from %s)", click.style(str(new_version), fg='blue'), click.style(str(ctx.obj.version), fg='blue'))
                version_bumped = True
                ctx.obj.version = new_version

                if 'file' in version_info:
                    replace_version(ctx.obj.config_dir / version_info['file'], ctx.obj.version)
                    repo.index.add((relative_version_file,))
                    if bump_message is not None:
                        commit_params.setdefault('message', bump_message)
        else:
            log.info("Skip version bumping due to the configuration or the target branch is not allowed to publish")

        commit_params.setdefault('author', author)
        commit_params.setdefault('committer', committer)
        if author_date is not None:
            commit_params['author_date'] = to_git_time(author_date)
        if commit_date is not None:
            commit_params['commit_date'] = to_git_time(commit_date)

        if bump_message is not None and not version_bumped:
            log.error("Version bumping requested, but the version policy '%s' decided not to bump from '%s'", bump['policy'], ctx.obj.version)
            return

        notes_ref = None
        if 'message' in commit_params:
            submit_commit = repo.index.commit(**commit_params)

            pkgs = utils.installed_pkgs()
            if pkgs and target_ref:
                notes_ref = f"refs/notes/hopic/{target_ref}"
                env = {
                        'GIT_AUTHOR_NAME': author.name,
                        'GIT_AUTHOR_EMAIL': author.email,
                        'GIT_COMMITTER_NAME': committer.name,
                        'GIT_COMMITTER_EMAIL': committer.email,
                    }
                if 'author_date' in commit_params:
                    env['GIT_AUTHOR_DATE'] = commit_params['author_date']
                if 'commit_date' in commit_params:
                    env['GIT_COMMITTER_DATE'] = commit_params['commit_date']

                hopic_commit_version = f"Committed-by: Hopic {utils.get_package_version(PACKAGE)}"
                notes_message = dedent("""\
                {hopic_commit_version}

                With Python version: {python_version}

                And with these installed packages:
                {pkgs}
                """).format(
                    hopic_commit_version=hopic_commit_version,
                    pkgs=pkgs,
                    python_version=platform.python_version())

                try:
                    notes = repo.git.notes('show', submit_commit.hexsha, ref=notes_ref)
                    if hopic_commit_version not in notes:
                        raise GitNotesMismatchError(submit_commit.hexsha, notes_message, notes)
                except git.GitCommandError:
                    notes = None

                if notes is None:
                    repo.git.notes(
                        'add', submit_commit.hexsha,
                        '--message=' + notes_message,
                        ref=notes_ref, env=env)

                notes_ref = f"{repo.commit(notes_ref)}:refs/notes/hopic/{target_ref}"
        else:
            submit_commit = repo.head.commit
        click.echo(submit_commit)

        autosquash_commits = [
                commit
                for commit in source_commits
                if commit.needs_autosquash()
            ]

        # Autosquash the merged commits (if any) to discover how that would look like.
        autosquash_base = None
        if autosquash_commits:
            commit = autosquash_commits[0]
            log.debug("Found an autosquash-commit in the source commits: '%s': %s", commit.subject, click.style(commit.hexsha, fg='yellow'))
            autosquash_base = repo.merge_base(target_commit, source_commit)
        autosquashed_commit = None
        if autosquash_base:
            repo.head.reference = source_commit
            repo.head.reset(index=True, working_tree=True)
            try:
                try:
                    env = {'GIT_SEQUENCE_EDITOR': ':'}
                    if 'commit_date' in commit_params:
                        env['GIT_COMMITTER_DATE'] = commit_params['commit_date']
                    repo.git.rebase(autosquash_base, interactive=True, autosquash=True, env=env, kill_after_timeout=300)
                except git.GitCommandError as e:
                    log.warning('Failed to perform auto squashing rebase: %s', e)
                else:
                    autosquashed_commit = repo.head.commit
                    if log.isEnabledFor(logging.DEBUG):
                        log.debug('Autosquashed to:')
                        for commit in git.Commit.list_items(repo, f"{target_commit}..{autosquashed_commit}", first_parent=True, no_merges=True):
                            subject = commit.message.splitlines()[0]
                            log.debug('%s %s', click.style(str(commit), fg='yellow'), subject)
            finally:
                repo.head.reference = submit_commit
                repo.head.reset(index=True, working_tree=True)

        update_submodules(repo, code_clean)

        if code_fresh:
            # Only restore mtimes when doing a clean build or working on a fresh clone. This prevents problems with timestamp-based build sytems.
            # I.e. make and ninja and probably half the world.
            restore_mtime_from_git(repo)

        # Tagging after bumping the version
        tagref = None
        version_tag = version_info.get('tag', False)
        if version_bumped and _is_valid_hotfix_base(ctx.obj.version) and version_tag and is_publish_allowed:
            if not isinstance(version_tag, str):
                assert ctx.obj.version is not None
                version_tag = ctx.obj.version.default_tag_name
            tagname = version_tag.format(
                    version        = ctx.obj.version,                                           # noqa: E251 "unexpected spaces around '='"
                    build_sep      = ('+' if getattr(ctx.obj.version, 'build', None) else ''),  # noqa: E251 "unexpected spaces around '='"
                )
            if hotfix and "-hotfix." not in tagname:
                raise VersioningError(
                    f"Tag '{tagname}' for hotfix version '{ctx.obj.version}' does not contain 'hotfix' in its prerelease portion.\n"
                    f"Likely the tag pattern ('{version_tag}') omitted the prerelease portion"
                )
            if 'build' in version_info and '+' not in tagname:
                tagname += f"+{version_info['build']}"
            tagref = repo.create_tag(
                tagname,
                submit_commit,
                force=True,
                message=f"Tagged-by: Hopic {utils.get_package_version(PACKAGE)}",
                env={
                    "GIT_COMMITTER_NAME": committer.name,
                    "GIT_COMMITTER_EMAIL": committer.email,
                },
            )

        # Re-read version to ensure that the newly created tag is taken into account
        ctx.obj.version, _ = determine_version(version_info, ctx.obj.config_dir, ctx.obj.code_dir)

        log.info('%s', repo.git.show(submit_commit, format='fuller', stat=True, notes='*'))

        push_commit = submit_commit
        if (ctx.obj.version is not None
                and 'file' in version_info and 'bump' in version_info.get('after-submit', {})
                and is_publish_allowed and bump['on-every-change']):
            params = {'bump': version_info['after-submit']['bump']}
            try:
                params['prerelease_seed'] = version_info['after-submit']['prerelease-seed']
            except KeyError:
                pass
            after_submit_version = ctx.obj.version.next_version(**params)
            log.debug("bumped post-submit version to: %s", click.style(str(after_submit_version), fg='blue'))

            new_version_file = StringIO()
            replace_version(ctx.obj.config_dir / version_info['file'], after_submit_version, outfile=new_version_file)
            new_version_data = new_version_file.getvalue().encode(sys.getdefaultencoding())

            old_version_blob = submit_commit.tree[relative_version_file]
            new_version_blob = git.Blob(
                repo=repo,
                binsha=repo.odb.store(gitdb.IStream(git.Blob.type, len(new_version_data), BytesIO(new_version_data))).binsha,
                mode=old_version_blob.mode,
                path=old_version_blob.path,
            )
            new_index = repo.index.from_tree(repo, submit_commit)
            new_index.add([new_version_blob], write=False)

            commit_params['author'] = commit_params['committer']
            if 'author_date' in commit_params and 'commit_date' in commit_params:
                commit_params['author_date'] = commit_params['commit_date']
            commit_params['message'] = f"[ Release build ] new version commit: {after_submit_version}\n"
            commit_params['parent_commits'] = (submit_commit,)
            # Prevent advancing HEAD
            commit_params['head'] = False

            push_commit = new_index.commit(**commit_params)
            log.info('%s', repo.git.show(push_commit, format='fuller', stat=True))

        refspecs = []
        bundle_excludes = [
            target_commit,
        ]
        bundle_refs: List[Tuple[str, Any]] = []

        if target_ref is not None:
            refspecs.append(f"{push_commit}:{target_ref}")
            bundle_refs.append((target_ref, push_commit))
        if tagref is not None:
            refspecs.append(f"{tagref.object}:{tagref.path}")
            bundle_refs.append((tagref.path, tagref.object))
        if notes_ref is not None:
            refspecs.append(notes_ref)
            notes_commit_name, notes_ref_name = notes_ref.split(":", 1)
            notes_commit = repo.commit(notes_commit_name)
            bundle_excludes.extend(notes_commit.parents)
            bundle_refs.append((notes_ref_name, notes_commit))

        commit_meta = {
            "remote": target_remote,
            "version-bumped": version_bumped,
            "ref": target_ref,
            "refspecs": refspecs,
            "target-commit": str(target_commit) if target_commit is not None else None,
            "source-commit": str(source_commit) if source_commit is not None else None,
            "autosquashed-commit": str(autosquashed_commit) if autosquashed_commit is not None else None,
        }

        store_commit_meta(repo, commit_meta, commit=submit_commit, old_commit=target_commit)

        if bundle is not None and bundle_refs:
            bundle_names = []

            bundle_ref_dir = Path(repo.git_dir) / "refs" / "bundle"
            for ref, obj in bundle_refs:
                if not ref.startswith("refs/") and not ref.startswith("heads/") and not ref.startswith("tags/"):
                    ref = f"heads/{ref}"
                elif ref.startswith("refs/"):
                    ref = ref[len("refs/") :]
                ref_path = bundle_ref_dir / ref
                ref_path.parent.mkdir(parents=True, exist_ok=True)
                ref_path.write_text(str(obj), encoding="UTF-8")
                bundle_names.append(f"refs/bundle/{ref}")

            if autosquashed_commit is not None:
                autosquashed_ref = Path(repo.git_dir) / "refs" / "hopic" / "bundle" / "autosquashed"
                autosquashed_ref.parent.mkdir(parents=True, exist_ok=True)
                autosquashed_ref.write_text(str(autosquashed_commit), encoding="UTF-8")
                bundle_names.append("refs/hopic/bundle/autosquashed")

            commit_meta_data = json.dumps(commit_meta).encode("UTF-8")
            commit_meta_blob = git.Blob(
                repo=repo,
                binsha=repo.odb.store(gitdb.IStream(git.Blob.type, len(commit_meta_data), BytesIO(commit_meta_data))).binsha,
                mode=0o100644,
                path="hopic-meta",
            )

            meta_index = repo.index.new(repo)
            meta_index.add([commit_meta_blob], write=False)

            commit_params["message"] = f"Hopic meta data for {submit_commit}"
            commit_params["parent_commits"] = ()
            # Prevent advancing HEAD
            commit_params["head"] = False

            meta_commit = meta_index.commit(
                **commit_params,
            )
            meta_ref = Path(repo.git_dir) / "refs" / "hopic" / "bundle" / "meta"
            meta_ref.parent.mkdir(parents=True, exist_ok=True)
            meta_ref.write_text(str(meta_commit), encoding="UTF-8")
            bundle_names.append("refs/hopic/bundle/meta")

            repo.git.bundle("create", bundle, *bundle_excludes, *bundle_names)

        if ctx.obj.version is not None:
            click.echo(ctx.obj.version)


@overload
def is_version_bump_enabled(bump_config: Mapping, ctx: Optional[click.Context] = None, *, is_publish_from_branch_allowed: bool) -> bool: ...


@overload
def is_version_bump_enabled(bump_config: Mapping, ctx: click.Context, *, is_publish_from_branch_allowed: Optional[bool] = None) -> bool: ...


def is_version_bump_enabled(bump_config: Mapping, ctx: Optional[click.Context] = None, *, is_publish_from_branch_allowed: Optional[bool] = None) -> bool:
    """
    Check if current branch with version configuration is allowed to bump version
    To avoid multiple is_publish_branch calls this can optionally be passed to this function
    """

    assert ctx is not None or is_publish_from_branch_allowed is not None
    if is_publish_from_branch_allowed is None:
        is_publish_from_branch_allowed = is_publish_branch(ctx)
    if not is_publish_from_branch_allowed:
        log.debug("not bumping the version because publishing from the current branch is disabled")
        return False
    if bump_config["policy"] == "disabled":
        log.debug("not bumping the version because the policy is configured to be disabled")
        return False
    if not bump_config["on-every-change"]:
        log.debug("not bumping the version because Hopic is configured not to do so on every change")
        return False
    return True


def get_current_version(ctx: click.Context) -> Version:
    version_info = ctx.obj.config["version"]
    if ctx.obj.version is None:
        if "file" in version_info:
            raise VersioningError(f"Failed to read the current version (from {version_info['file']}) while attempting to bump the version")

        msg = "Failed to determine the current version while attempting to bump the version"
        log.error(msg)
        # TODO: PIPE-309: provide an initial starting point instead
        log.info("If this is a new repository you may wish to create a 0.0.0 tag for Hopic to start bumping from")
        raise VersioningError(msg)

    return ctx.obj.version


def parse_commit_range(
    repo: git.Repo,
    from_commit: Union[None, git.objects.commit.Commit, str],
    to_commit: Union[None, git.objects.commit.Commit, str],
    bump_config: Mapping,
) -> Iterable[CommitMessage]:
    if from_commit is None or to_commit is None:
        return

    for commit in git.Commit.list_items(
        repo,
        f"{from_commit}..{to_commit}",
        first_parent=bump_config.get("first-parent", True),
        no_merges=bump_config.get("no-merges", True),
    ):
        yield parse_commit_message(commit, policy=bump_config["policy"], strict=bump_config.get("strict", False))


@prepare_source_tree.command()
@click.pass_context
# git
@click.option('--source-remote' , metavar='<url>', help='<source> remote to merge into <target>')
@click.option('--source-ref'    , metavar='<ref>', help='ref of <source> remote to merge into <target>')
@click.option('--change-request', metavar='<identifier>'           , help='Identifier of change-request to use in merge commit message')
@click.option('--title'         , metavar='<title>'                , help='''Change request title to incorporate in merge commit's subject line''')
@click.option('--description'   , metavar='<description>'          , help='''Change request description to incorporate in merge commit message's body''')
@click.option('--approved-by'   , metavar='<approver>'             , help='''Name of approving reviewer (can be provided multiple times).''', multiple=True)
def merge_change_request(
            ctx,
            source_remote,
            source_ref,
            change_request,
            title,
            description,
            approved_by,
        ):
    """
    Merges the change request from the specified branch.
    """

    def get_valid_approvers(repo, approved_by_list, source_remote, source_commit):
        """Inspects approvers list and, where possible, checks if approval is still valid."""

        valid_hash_re = re.compile(r"^(.+):([0-9a-zA-Z]{40})$")
        autosquash_re = re.compile(r'^(fixup|squash)!\s+')
        valid_approvers = []

        # Fetch the hashes from the remote in one go
        approved_hashes = [entry.group(2) for entry in (valid_hash_re.match(entry) for entry in approved_by_list) if entry]
        try:
            source_remote.fetch(approved_hashes)
        except git.GitCommandError:
            log.warning("One or more of the last reviewed commit hashes invalid: '%s'", ' '.join(approved_hashes))

        for approval_entry in approved_by_list:
            hash_match = valid_hash_re.match(approval_entry)
            if not hash_match:
                valid_approvers.append(approval_entry)
                continue

            approver, last_reviewed_commit_hash = hash_match.groups()
            try:
                last_reviewed_commit = repo.commit(last_reviewed_commit_hash)
            except ValueError:
                log.warning("Approval for '%s' is ignored, as the associated hash is unknown or invalid: '%s'", approver, last_reviewed_commit_hash)
                continue

            if last_reviewed_commit_hash == source_commit.hexsha:
                valid_approvers.append(approver)
                continue
            if last_reviewed_commit.diff(source_commit):
                log.warning(
                        "Approval for '%s' is not valid anymore due to content changes compared to last reviewed commit '%s'",
                        approver, last_reviewed_commit_hash)
                continue

            # Source has a different hash, but no content diffs.
            # Now 'squash' and compare metadata (author, date, commit message).
            merge_base = repo.merge_base(repo.head.commit, source_commit)

            source_commits = [
                    (commit.author, commit.authored_date, commit.message.rstrip()) for commit in
                    git.Commit.list_items(repo, merge_base[0].hexsha + '..' + source_commit.hexsha, first_parent=True, no_merges=True)]

            autosquashed_reviewed_commits = [
                    (commit.author, commit.authored_date, commit.message.rstrip()) for commit in
                    git.Commit.list_items(repo, merge_base[0].hexsha + '..' + last_reviewed_commit.hexsha, first_parent=True, no_merges=True)
                    if not autosquash_re.match(commit.message)]

            log.debug(
                    "For approver '%s', checking source commits:\n%s\n.. against squashed reviewed commits:\n%s",
                    approver, source_commits, autosquashed_reviewed_commits)

            if autosquashed_reviewed_commits == source_commits:
                log.debug("Approval for '%s' is still valid", approver)
                valid_approvers.append(approver)
            else:
                log.warning(
                        "Approval for '%s' is not valid anymore due to metadata changes compared to last reviewed commit '%s'",
                        approver, last_reviewed_commit_hash)
        return valid_approvers

    def change_applicator(repo, author, committer):
        try:
            source = repo.remotes.source
        except AttributeError:
            source = repo.create_remote('source', source_remote)
        else:
            source.set_url(source_remote)
        source_commit = source.fetch(source_ref)[0].commit

        repo.git.merge(source_commit, no_ff=True, no_commit=True, env={
            'GIT_AUTHOR_NAME': author.name,
            'GIT_AUTHOR_EMAIL': author.email,
            'GIT_COMMITTER_NAME': committer.name,
            'GIT_COMMITTER_EMAIL': committer.email,
        })

        msg = f"Merge #{change_request}"
        if title is not None:
            msg = f"{msg}: {title}\n"
        if description is not None:
            msg = f"{msg}\n{description}\n"

        # Prevent splitting footers with empty lines in between, because 'git interpret-trailers' doesn't like it.
        parsed_msg = parse_commit_message(msg)
        if not parsed_msg.footers:
            msg += u'\n'

        approvers = get_valid_approvers(repo, approved_by, source, source_commit)
        if approvers:
            msg += '\n'.join(f"Acked-by: {approver}" for approver in approvers) + u'\n'
        msg += f'Merged-by: Hopic {get_package_version(PACKAGE)}\n'

        # Reread config & install extensions after potential configuration file change
        assert hasattr(ctx.obj, "pip_constraints")
        install_extensions_and_parse_config(ctx.obj.pip_constraints)

        bump = ctx.obj.config['version']['bump']
        strict = bump.get('strict', False)
        try:
            merge_commit = parse_commit_message(msg, policy=bump['policy'], strict=strict)
        except Exception as e:
            if bump['policy'] == 'conventional-commits':
                log.error(
                    "The pull request title could not be parsed as a conventional commit.\n"
                    "Parsing the PR title failed due to:\n%s",
                    "".join(f" - {problem}\n" for problem in str(e).split('\n'))
                )
                ctx.exit(1)
            raise

        if is_version_bump_enabled(bump, ctx) and strict:
            source_commits = parse_commit_range(repo, repo.head.commit, source_commit, bump)
            base_version = get_current_version(ctx)
            new_version = base_version.next_version_for_commits(source_commits)
            merge_commit_next_version = base_version.next_version_for_commits([merge_commit])
            if new_version != merge_commit_next_version:
                raise VersionBumpMismatchError(new_version, merge_commit_next_version)

        return {
                'config_parsed': True,
                'message': msg,
                'parent_commits': (
                    repo.head.commit,
                    source_commit,
                ),
                'source_commit': source_commit,
            }
    return change_applicator


@prepare_source_tree.command()  # noqa: E302 'expected 2 blank lines'
@click.argument('modality', autocompletion=autocomplete.modality_from_config)
@click.pass_context
def apply_modality_change(
            ctx,
            modality,
        ):
    """
    Applies the changes specific to the specified modality.
    """

    def change_applicator(repo, author, committer):
        assert hasattr(ctx.obj, "pip_constraints")
        install_extensions_and_parse_config(ctx.obj.pip_constraints)

        modality_cmds = ctx.obj.config["modality-source-preparation"][modality]

        has_changed_files = any("changed-files" in cmd for cmd in modality_cmds)

        (commit_message,) = (
            cmd.get("commit-message", cmd.get("commit-message-cmd")) for cmd in modality_cmds if "commit-message" in cmd or "commit-message-cmd" in cmd
        )

        if not has_changed_files:
            # Force clean builds when we don't know how to discover changed files
            repo.git.clean(x=True, d=True, force=True)

        volume_vars = ctx.obj.volume_vars.copy()
        volume_vars.setdefault('HOME', os.path.expanduser('~'))
        vars_from_env = {key: value for key, value in os.environ.items() if key in ctx.obj.config["pass-through-environment-vars"]}
        vars_from_env.update(volume_vars)
        volume_vars = vars_from_env

        # Set submit_commit to None to indicate that we haven't got a submittable commit (yet).
        hopic_git_info = HopicGitInfo.from_repo(repo)._replace(submit_commit=None)
        (*_,) = build.build_variant(variant=modality, cmds=modality_cmds, hopic_git_info=hopic_git_info, exec_stdout=sys.__stderr__, cwd="${CFGDIR}")

        if not has_changed_files:
            # 'git add --all' equivalent (excluding the code_dir)
            add_files = set(repo.untracked_files)
            remove_files = set()
            with repo.config_reader() as cfg:
                try:
                    code_dir = cfg.get_value('hopic.code', 'dir')
                except (NoOptionError, NoSectionError):
                    pass
                else:
                    if code_dir in add_files:
                        add_files.remove(code_dir)
                    if (code_dir + '/') in add_files:
                        add_files.remove(code_dir + '/')

            for diff in repo.index.diff(None):
                if not diff.deleted_file:
                    add_files.add(diff.b_path)
                remove_files.add(diff.a_path)
            remove_files -= add_files
            if remove_files:
                repo.index.remove(remove_files)
            if add_files:
                repo.index.add(add_files)

        if not repo.index.diff(repo.head.commit):
            log.info("No changes introduced by '%s'", modality)
            return None

        if isinstance(commit_message, str):
            commit_message = expand_vars(volume_vars, commit_message)
        else:
            assert isinstance(commit_message, Mapping)
            (commit_message,) = build.build_variant(
                variant=modality, cmds=[commit_message], hopic_git_info=hopic_git_info, exec_stdout=subprocess.PIPE, cwd="${CFGDIR}"
            )

        if commit_message[-1:] != "\n":
            commit_message += "\n"

        # Prevent splitting footers with empty lines in between, because 'git interpret-trailers' doesn't like it.
        parsed_msg = parse_commit_message(commit_message)
        if not parsed_msg.footers:
            commit_message += "\n"

        commit_message += f"Merged-by: Hopic {utils.get_package_version(PACKAGE)}\n"

        commit_params = {'message': commit_message}
        # If this change was a merge make sure to produce a merge commit for it
        try:
            commit_params['parent_commits'] = (
                    repo.commit('ORIG_HEAD'),
                    repo.commit('MERGE_HEAD'),
                )
        except git.BadName:
            pass
        return commit_params

    return change_applicator


@prepare_source_tree.command()
@click.pass_context
def bump_version(ctx):
    """
    Bump the version based on the configuration.
    """

    def change_applicator(repo, author, committer):
        gitversion = determine_git_version(repo)
        if gitversion.exact:
            log.info("Not bumping because no new commits are present since the last tag '%s'", gitversion.tag_name)
            return None
        tag = repo.tags[gitversion.tag_name]
        return {
            'bump_message': dedent(f"""\
                    chore: release new version

                    Bumped-by: Hopic {utils.get_package_version(PACKAGE)}
                    """),
            'base_commit': tag.commit,
            'bump-override': {
                'on-every-change': True,
                'strict': False,
                'first-parent': False,
                'no-merges': False,
            },
        }

    return change_applicator


@main.command()
# fmt: off
@click.option('--phase'      , '-p' , metavar='<phase>'  , multiple=True, help='''Build phase''', autocompletion=autocomplete.phase_from_config)
@click.option('--variant'    , '-v' , metavar='<variant>', multiple=True, help='''Configuration variant''', autocompletion=autocomplete.variant_from_config)
@click.option("--modality"   , "-m" , metavar="<modality>",               help='''Display only meta-data for the specified modality.''')
@click.option('--post-submit'       , is_flag=True       ,                help='''Display only post-submit meta-data.''')
# fmt: on
@click.pass_context
def getinfo(
    ctx: click.Context,
    phase: Sequence[str],
    variant: Sequence[str],
    modality: Optional[str],
    post_submit: Optional[str],
):
    """
    Display meta-data associated with each (or the specified) variant in each (or the specified) phase.

    The output is JSON encoded.

    If a phase or variant filter is specified the name of that will not be present in the output.
    Otherwise this is a nested dictionary of phases and variants.
    """
    info: Dict[str, Any] = OrderedDict()

    def append_meta_from_cmd(info, cmd: typing.Mapping[str, Any], permitted_fields: Set):
        assert isinstance(cmd, Mapping)

        info = info.copy()

        for key, val in cmd.items():
            if key == "timeout" and "sh" not in cmd:
                # Return _global_ timeout (not "sh"-specific) only
                pass
            elif key == "finally":
                for final_cmd in cmd["finally"]:
                    info.update(append_meta_from_cmd(info, final_cmd, permitted_fields))
                continue
            elif key not in permitted_fields:
                continue

            try:
                val = expand_vars(ctx.obj.volume_vars, val)
            except KeyError:
                pass
            else:
                if key == "timeout":
                    if key in info:
                        info[key] += val
                    else:
                        info[key] = val
                elif isinstance(info.get(key), Mapping):
                    assert isinstance(info[key], MutableMapping)
                    for subkey, subval in val.items():
                        info[key].setdefault(subkey, subval)
                elif isinstance(info.get(key), MutableSequence):
                    info[key].extend(val)
                else:
                    info.setdefault(key, val)

        return info

    if modality and post_submit:
        log.error("--modality and --post-submit are mutually exclusive options")
        ctx.exit(1)
    if phase or variant:
        if modality:
            log.error("--modality is mutually exclusive with --phase and --variant")
            ctx.exit(1)
        if post_submit:
            log.error("--post-submit is mutually exclusive with --phase and --variant")
            ctx.exit(1)

    permitted_fields: AbstractSet[str]
    if modality:
        permitted_fields = {
            "with-credentials",
        }
        for cmd in ctx.obj.config["modality-source-preparation"].get(modality, ()):
            info.update(append_meta_from_cmd(info, cmd, permitted_fields))
            if "commit-message-cmd" in cmd:
                info.update(append_meta_from_cmd(info, cmd["commit-message-cmd"], permitted_fields))
    elif post_submit:
        permitted_fields = frozenset({
            'node-label',
            'with-credentials',
        })
        for phasename, cmds in ctx.obj.config['post-submit'].items():
            for cmd in cmds:
                info.update(append_meta_from_cmd(info, cmd, permitted_fields))
    else:
        permitted_fields = frozenset({
            'archive',
            'wait-on-full-previous-phase',
            'fingerprint',
            'junit',
            'node-label',
            'run-on-change',
            'stash',
            'with-credentials',
            'worktrees',
        })
        for phasename, curphase in ctx.obj.config['phases'].items():
            if phase and phasename not in phase:
                continue
            for variantname, curvariant in curphase.items():
                if variant and variantname not in variant:
                    continue

                # Only store phase/variant keys if we're not filtering on a single one of them.
                var_info = info
                if len(phase) != 1:
                    var_info = var_info.setdefault(phasename, OrderedDict())
                if len(variant) != 1:
                    var_info = var_info.setdefault(variantname, OrderedDict())

                for cmd in curvariant:
                    var_info.update(append_meta_from_cmd(var_info, cmd, permitted_fields))

                # mark empty variants as being a nop
                if (
                    len(curvariant) == 0
                    or (
                        len(curvariant) == 1
                        and dict(curvariant[0]) == {"wait-on-full-previous-phase": False}
                    )
                ):
                    var_info["nop"] = True
    click.echo(json.dumps(info, indent=4, separators=(',', ': '), cls=JSONEncoder))


@main.command()
@click.argument("bundle", type=click.Path(file_okay=True, dir_okay=False, readable=True))
@click.pass_context
def unbundle(ctx, *, bundle: PathLike):
    """
    Unbundle the specified git bundle and setup all included refs for pushing.
    """

    with git.Repo(ctx.obj.workspace) as repo:
        target_commit = repo.head.commit

        with repo.config_reader() as cfg:
            section = f"hopic.{target_commit}"
            target_ref = cfg.get(section, "ref", fallback=None)
            target_remote = cfg.get(section, "remote", fallback=None)
            code_clean = cfg.getboolean("hopic.code", "cfg-clean", fallback=False)
            code_fresh = cfg.getboolean("hopic.code", "fresh", fallback=code_clean)

        submit_commit: Optional[git.Commit] = None
        commit_meta: Dict[str, Any] = {}
        bundle_path = "refs/bundle/"
        head_path = "heads/"
        unbundle_proc = repo.git.bundle("unbundle", bundle, as_process=True)
        for headline in unbundle_proc.stdout:
            obj, ref = headline.decode("UTF-8").rstrip("\n").split(" ", 1)
            if ref == "refs/hopic/bundle/meta":
                meta_commit = repo.commit(obj)

                m = re.match(r"^Hopic meta\s*data for \b([0-9a-fA-F]+)\b", meta_commit.message)
                if not m:
                    continue
                submit_commit = repo.commit(m.group(1))

                if "hopic-meta" not in meta_commit.tree:
                    continue
                commit_meta = json.load(meta_commit.tree["hopic-meta"].data_stream)
                store_commit_meta(repo, commit_meta, commit=submit_commit, old_commit=target_commit)
            if not ref.startswith(bundle_path):
                continue
            ref = ref[len(bundle_path) :]
            if not ref.startswith(head_path):
                ref_path = Path(repo.git_dir) / "refs" / ref
                ref_path.parent.mkdir(parents=True, exist_ok=True)
                ref_path.write_text(str(obj), encoding="UTF-8")
        try:
            unbundle_proc.terminate()
        except OSError:
            pass

    if not submit_commit or not commit_meta:
        log.error("Couldn't find Hopic meta data inside given bundle")
        ctx.exit(1)

    workspace = ctx.obj.workspace
    # Check out specified repository
    checkout_tree(
        ctx.obj.workspace,
        remote=None,
        ref=commit_meta["ref"],
        commit=submit_commit,
        clean=code_clean,
    )

    try:
        ctx.obj.config = read_config(determine_config_file_name(ctx), ctx.obj.volume_vars)
        with git.Repo(workspace) as repo:
            if code_clean:
                clean_repo(repo, ctx.obj.config["clean"])
            if code_fresh:
                # Only restore mtimes when doing a clean build or working on a fresh clone. This prevents problems with timestamp-based build sytems.
                # I.e. make and ninja and probably half the world.
                restore_mtime_from_git(repo)
        git_cfg = ctx.obj.config["scm"]["git"]
    except (click.BadParameter, KeyError, TypeError, OSError, IOError, YAMLError):
        return
    finally:
        log.info("%s", repo.git.show(submit_commit, format="fuller", stat=True, notes="*"))

    checkout_worktrees(workspace, git_cfg["worktrees"])

    if "remote" not in git_cfg and "ref" not in git_cfg:
        return

    code_dir = find_code_dir(workspace)

    # Check out configured repository and mark it as the code directory of this one
    with git.Repo(workspace) as repo, repo.config_writer() as cfg:
        cfg.remove_section("hopic.code")
        cfg.set_value("hopic.code", "dir", str(code_dir))
        cfg.set_value("hopic.code", "cfg-remote", target_remote)
        cfg.set_value("hopic.code", "cfg-ref", target_ref)
        cfg.set_value("hopic.code", "cfg-clean", str(code_clean))
        if code_fresh:
            cfg.set_value("hopic.code", "fresh", str(code_fresh))

    checkout_tree(
        code_dir,
        git_cfg.get("remote", target_remote),
        git_cfg.get("ref", target_ref),
        clean=code_clean,
        clean_config=ctx.obj.config["clean"],
    )


@main.command()
@click.option('--bundle', metavar='<file>', help='Git bundle to use', type=click.Path(file_okay=True, dir_okay=False, readable=True, resolve_path=True))
@click.pass_context
def unbundle_worktrees(ctx, bundle):
    """
    Unbundle a git bundle and fast-forward all the configured worktrees that are included in it.
    """

    with git.Repo(ctx.obj.workspace) as repo:
        submit_commit = repo.head.commit
        section = f"hopic.{submit_commit}"
        with repo.config_reader() as git_cfg:
            try:
                refspecs = list(shlex.split(git_cfg.get_value(section, 'refspecs')))
            except (NoOptionError, NoSectionError):
                refspecs = []

        head_path = 'refs/heads/'
        worktrees = dict((v, k) for k, v in ctx.obj.config['scm']['git']['worktrees'].items())
        for headline in repo.git.bundle('list-heads', bundle).splitlines():
            commit, ref = headline.split(' ', 1)
            if not ref.startswith(head_path):
                continue
            ref = ref[len(head_path):]
            if ref not in worktrees:
                continue

            subdir = worktrees[ref]
            log.debug("Checkout worktree '%s' to '%s' (proposed branch '%s')", subdir, commit, ref)
            checkout_tree(
                ctx.obj.workspace / subdir,
                bundle,
                ref,
                remote_name="bundle",
                tags=False,
            )
            refspecs.append(f"{commit}:{ref}")

        # Eliminate duplicate pushes to the same ref and replace it by a single push to the _last_ specified object
        seen_refs = set()
        new_refspecs = []
        for refspec in reversed(refspecs):
            _, ref = refspec.rsplit(':', 1)
            if ref in seen_refs:
                continue
            new_refspecs.insert(0, refspec)
            seen_refs.add(ref)
        refspecs = new_refspecs

        with repo.config_writer() as cfg:
            cfg.set_value(section, 'refspecs', ' '.join(shlex.quote(refspec) for refspec in refspecs))


@main.command()
@click.option('--target-remote', metavar='<url>', help='''The remote to push to, if not specified this will default to the checkout remote.''')
@click.pass_context
def submit(ctx, target_remote):
    """
    Submit the changes created by prepare-source-tree to the target remote.
    """

    with git.Repo(ctx.obj.workspace) as repo:
        hopic_git_info = HopicGitInfo.from_repo(repo)
        if target_remote is None:
            target_remote = hopic_git_info.submit_remote

        repo.git.push(target_remote, hopic_git_info.refspecs, atomic=True)

        with repo.config_writer() as cfg:
            cfg.remove_section(f"hopic.{repo.head.commit}")

    for phase in ctx.obj.config['post-submit'].values():
        (*_,) = build.build_variant(variant="post-submit", cmds=phase, hopic_git_info=hopic_git_info)


@main.command()
@click.pass_context
def show_config(ctx):
    """
    Diagnostic helper command to display the configuration after processing.
    """

    click.echo(json.dumps(ctx.obj.config, indent=4, separators=(',', ': '), cls=JSONEncoder))


@main.command()
@click.pass_context
def show_env(ctx):
    """
    Diagnostic helper command to display the execution environment.
    """

    click.echo(json.dumps(ctx.obj.volume_vars, indent=4, separators=(',', ': '), sort_keys=True))



# Copyright (c) 2019 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent
from typing import Optional

from click import ClickException


class ConfigurationError(ClickException):
    exit_code = 32

    def __init__(self, message, file=None):
        super().__init__(message)
        self.file = file

    def format_message(self):
        if self.file is not None:
            return "configuration error in '%s': %s" % (self.file, self.message)
        else:
            return "configuration error: %s" % (self.message,)


class VersioningError(ClickException):
    exit_code = 33


class MissingCredentialVarError(ClickException):
    exit_code = 34

    def __init__(self, credential_id, var_name):
        super().__init__(f"credential '{credential_id}' not available when trying to expand variable '{var_name}'")
        self.credential_id = credential_id
        self.var_name      = var_name


class UnknownPhaseError(ClickException):
    exit_code = 35

    def __init__(self, phase):
        super().__init__(f"build does not contain phase(s): {', '.join(phase)}")
        self.phase = phase


class VersionBumpMismatchError(ClickException):
    exit_code = 36

    def __init__(self, commit_version, merge_version):
        super().__init__(f"Version bump for commit messages results in different version ({commit_version}) "
                         f"than the version based on the merge message ({merge_version}).")


class CommitAncestorMismatchError(ClickException):
    exit_code = 37

    def __init__(self, commit, ancestor_commit, ref):
        super().__init__(
            dedent(
                """\
                attempting to checkout commit '{self.commit}' which is not an ancestor of remote ref '{self.ref}' ('{self.ancestor_commit}')
                possibly remote ref '{self.ref}' was force pushed to
                """
            )
        )
        self.commit = commit
        self.ancestor_commit = ancestor_commit
        self.ref = ref

    def format_message(self):
        return self.message.format(self=self)

    def __str__(self):
        return self.format_message()


class MissingFileError(ClickException):
    exit_code = 38


class GitNotesMismatchError(ClickException):
    exit_code = 39

    def __init__(self, object, new_note, existing_note):
        super().__init__(
            dedent(
                """\
                attempting to add a different note to object '{self.object}' which already had a Hopic note
                new note:
                {self.new_note}

                existing note:
                {self.existing_note}
                """
            )
        )
        self.object = object
        self.new_note = new_note
        self.existing_note = existing_note

    def format_message(self):
        return self.message.format(self=self)

    def __str__(self):
        return self.format_message()


class StepTimeoutExpiredError(ClickException):
    exit_code = 40

    def __init__(self, timeout, *, cmd: Optional[str] = None, before: bool = False):
        msg = f"Timeout of {timeout} seconds expired {'before' if before else 'while'} executing build command"
        if cmd:
            msg += f": {cmd}"
        super().__init__(msg)
        self.timeout = timeout



# Copyright (c) 2019 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from textwrap import dedent

import git
import pytest


_git_time = f"{7 * 24 * 3600} +0000"
_author = git.Actor('Bob Tester', 'bob@example.net')
_commitargs = dict(
    author_date=_git_time,
    commit_date=_git_time,
    author=_author,
    committer=_author,
)


def test_clean_submodule_checkout(capfd, monkeypatch, run_hopic, tmp_path):
    author = git.Actor('Bob Tester', 'bob@example.net')
    commitargs = dict(
            author_date=_git_time,
            commit_date=_git_time,
            author=author,
            committer=author,
        )

    old_subcommand_getter = git.cmd.Git.__getattr__

    def new_subcommand_getter(self, name: str):
        if name == "submodule":
            self(c="protocol.file.allow=always")
        return old_subcommand_getter(self, name)

    monkeypatch.setattr("git.cmd.Git.__getattr__", new_subcommand_getter)

    dummy_content = 'Lalalala!\n'
    subrepo = tmp_path / 'subrepo'
    with git.Repo.init(subrepo, expand_vars=False) as repo:
        (subrepo / 'dummy.txt').write_text(dummy_content)
        repo.index.add(('dummy.txt',))
        repo.index.commit(message='Initial dummy commit', **commitargs)

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / 'hopic-ci-config.yaml').write_text('''\
phases:
  build:
    test:
      - cat subrepo/dummy.txt
''')
        repo.index.add(('hopic-ci-config.yaml',))
        repo.git.submodule(('add', '../subrepo', 'subrepo'))
        repo.index.commit(message='Initial commit', **commitargs)

    # Successful checkout and build
    (_, result) = run_hopic(
            ('checkout-source-tree', '--clean', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('build',),
        )
    assert result.exit_code == 0
    out, err = capfd.readouterr()
    build_out = ''.join(out.splitlines(keepends=True)[1:])
    assert build_out == dummy_content

    # Make submodule checkout fail
    subrepo.rename(subrepo.parent / 'old-subrepo')

    # Expected failure
    with pytest.raises(git.GitCommandError, match=r'(?is)submodule.*repository.*\bdoes not exist\b'):
        (_,) = run_hopic(('checkout-source-tree', '--clean', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'))

    # Ignore submodule failure only
    (result,) = run_hopic(
        ('checkout-source-tree', '--clean', '--ignore-initial-submodule-checkout-failure', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
    )
    assert result.exit_code == 0


def test_clean_checkout_in_non_empty_dir(run_hopic, tmp_path):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        repo.index.commit(message='Initial commit', **_commitargs)

    non_empty_dir = tmp_path / 'non-empty-clone'
    non_empty_dir.mkdir(parents=True)
    garbage_file = non_empty_dir / 'not-empty'
    garbage_file.write_text('Garbage!')

    # Verify first that a checkout without --clean fails
    with pytest.raises(git.GitCommandError, match=r'(?is).*\bfatal:\s+destination\s+path\b.*\bexists\b.*\bnot\b.*\bempty\s+directory\b'):
        (_,) = run_hopic(('--workspace', non_empty_dir, 'checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'))
    assert garbage_file.exists()

    # Now notice the difference and expect it to succeed with --clean
    (clean_result,) = run_hopic(
        ('--workspace', non_empty_dir, 'checkout-source-tree', '--clean', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
    )
    assert clean_result.exit_code == 0
    assert not garbage_file.exists()


def test_checkout_in_newly_initialized_repo(run_hopic, tmp_path):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        repo.index.commit(message='Initial commit', **_commitargs)

    new_init_repo = tmp_path / 'non-empty-clone'
    with git.Repo.init(new_init_repo, expand_vars=False):
        pass

    (result,) = run_hopic(('--workspace', new_init_repo, 'checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'))
    assert result.exit_code == 0


def test_default_clean_checkout_option(run_hopic):
    author = git.Actor('Bob Tester', 'bob@example.net')
    commitargs = dict(
        author_date=_git_time,
        commit_date=_git_time,
        author=author,
        committer=author,
    )

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / 'hopic-ci-config.yaml').write_text('''\
{}
''')
        temp_test_file = run_hopic.toprepo / "random_file.txt"
        temp_test_file.write_text('''\
nothing to see here
''')

        repo.index.add(('hopic-ci-config.yaml',))
        repo.index.commit(message='Initial commit', **commitargs)
        commit = list(repo.iter_commits('master', max_count=1))[0]
        (result,) = run_hopic(
            ('--workspace', run_hopic.toprepo, 'checkout-source-tree', '--clean', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
        )
        assert result.exit_code == 0
        assert not temp_test_file.is_file()
        assert commit.committed_date == (run_hopic.toprepo / 'hopic-ci-config.yaml').stat().st_mtime


def test_clean_option_custom_command_is_run_before_default_command(run_hopic):
    author = git.Actor('Bob Tester', 'bob@example.net')
    commitargs = dict(
        author_date=_git_time,
        commit_date=_git_time,
        author=author,
        committer=author,
    )

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / 'hopic-ci-config.yaml').write_text('''\
clean:
    - touch dummy.txt
''')
        temp_test_file = 'random_file.txt'
        (run_hopic.toprepo / temp_test_file).write_text('''\
nothing to see here
''')

        repo.index.add(('hopic-ci-config.yaml',))
        repo.index.commit(message='Initial commit', **commitargs)
        commit = list(repo.iter_commits('master', max_count=1))[0]
        (result,) = run_hopic(
            ('--workspace', run_hopic.toprepo, 'checkout-source-tree', '--clean', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
        )
        assert result.exit_code == 0
        assert not (run_hopic.toprepo / temp_test_file).is_file()
        assert not (run_hopic.toprepo / 'dummy.txt').is_file()
        assert commit.committed_date == (run_hopic.toprepo / 'hopic-ci-config.yaml').stat().st_mtime


def test_clean_option_custom_command_is_executed(capfd, run_hopic):
    author = git.Actor('Bob Tester', 'bob@example.net')
    commitargs = dict(
        author_date=_git_time,
        commit_date=_git_time,
        author=author,
        committer=author,
    )

    std_out_message = 'test dummy'
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / 'hopic-ci-config.yaml').write_text(
            dedent(
                f"""\
                clean:
                    - echo '{std_out_message}'
                """
            )
        )
        temp_test_file = 'random_file.txt'
        (run_hopic.toprepo / temp_test_file).write_text('''\
nothing to see here
''')

        repo.index.add(('hopic-ci-config.yaml',))
        repo.index.commit(message='Initial commit', **commitargs)
        commit = list(repo.iter_commits('master', max_count=1))[0]
        (result,) = run_hopic(
            ('--workspace', run_hopic.toprepo, 'checkout-source-tree', '--clean', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
        )
        assert result.exit_code == 0
        out, err = capfd.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)
        clean_out = out.splitlines()[0]
        assert clean_out == std_out_message
        assert not (run_hopic.toprepo / temp_test_file).is_file()
        assert commit.committed_date == (run_hopic.toprepo / 'hopic-ci-config.yaml').stat().st_mtime


def test_clean_option_home_annotations(capfd, run_hopic):
    author = git.Actor('Bob Tester', 'bob@example.net')
    commitargs = dict(
        author_date=_git_time,
        commit_date=_git_time,
        author=author,
        committer=author,
    )
    home_path = os.path.expanduser("~")
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / 'hopic-ci-config.yaml').write_text('''\
clean:
    - echo '$HOME'
    - echo '~'
''')

        repo.index.add(('hopic-ci-config.yaml',))
        repo.index.commit(message='Initial commit', **commitargs)
        (result,) = run_hopic(
            ('--workspace', run_hopic.toprepo, 'checkout-source-tree', '--clean', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
        )
        assert result.exit_code == 0
        out, err = capfd.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)
        clean_home_out = out.splitlines()[0]
        clean_tilde_out = out.splitlines()[1]
        assert clean_home_out == home_path
        assert clean_tilde_out == home_path


def test_handle_syntax_error_in_optional_hopic_file(run_hopic):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / 'hopic-ci-config.yaml').write_text(
            dedent(
                '''
                phases:
                  build:
                    test:
                      - image:
                        some-image-with-an-entrypoint:0.0.42
                      - "-o subrepo/dummy.txt""
                '''
            )
        )
        repo.index.add(('hopic-ci-config.yaml',))
        repo.index.commit(message='Commit incorrect hopic file', **_commitargs)

    # checkout-source-tree should be successful
    (result,) = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
        )
    assert result.exit_code == 0


def test_checkout_non_head_commit(run_hopic):
    dummy = run_hopic.toprepo / "dummy.txt"
    first_content = "Lalalala!\n"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        dummy.write_text(first_content)
        repo.index.add((str(dummy.relative_to(run_hopic.toprepo)),))
        first_commit = repo.index.commit(message="Initial dummy commit", **_commitargs)

        dummy.write_text("Mooh!\n")
        repo.index.add((str(dummy.relative_to(run_hopic.toprepo)),))
        repo.index.commit(message="Subsequent dummy commit", **_commitargs)

    (result,) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", "master", "--target-commit", first_commit),
    )
    assert result.exit_code == 0


def test_reject_checkout_out_of_branch_commit(run_hopic):
    dummy = run_hopic.toprepo / "dummy.txt"
    first_content = "Lalalala!\n"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        dummy.write_text(first_content)
        repo.index.add((str(dummy.relative_to(run_hopic.toprepo)),))
        first_commit = repo.index.commit(message="Initial dummy commit", **_commitargs)

        dummy.write_text("Mooh!\n")
        repo.index.add((str(dummy.relative_to(run_hopic.toprepo)),))
        final_commit = repo.index.commit(message="Subsequent dummy commit", **_commitargs)
        repo.heads.master.commit = first_commit
        repo.head.reference = repo.heads.master
        repo.head.reset(index=True, working_tree=True)

    (result,) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", "master", "--target-commit", final_commit),
    )
    assert result.exit_code == 37



# Copyright (c) 2018 - 2021 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import (
        datetime,
        timedelta,
    )
from enum import Enum, unique
import os
import logging
import math
import sys
from typing import (
    Any,
    Iterable,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from dateutil.tz import (
        tzlocal,
        tzutc,
    )
import git

from .types import (
    PathLike,
)
from .versioning import (
    GitVersion,
    Version,
    read_version,
)

log = logging.getLogger(__name__)


def to_git_time(date: datetime) -> str:
    """
    Converts a datetime object to a string with Git's internal time format.

    This is necessary because GitPython, wrongly, interprets an ISO-8601 formatted time string as
    UTC time to be converted to the specified timezone.

    Git's internal time format actually is UTC time plus a timezone to be applied for display
    purposes, so converting it to that yields correct behavior.
    """

    utctime = int((
        date - datetime.utcfromtimestamp(0).replace(tzinfo=tzutc())
    ).total_seconds())
    return f"{utctime} {date:%z}"


def determine_source_date(workspace: Union[git.Repo, PathLike]) -> Optional[datetime]:
    """
    Determine the date of most recent change to the sources in the given workspace
    """

    try:
        if not isinstance(workspace, git.Repo):
            repo = git.Repo(workspace)
        else:
            repo = workspace
    except (git.InvalidGitRepositoryError, git.NoSuchPathError):
        return None

    try:
        try:
            source_date = repo.head.commit.committed_datetime
        except ValueError:
            # This happens for a repository that has been initialized but for which a commit hasn't yet been created
            # or checked out.
            #     $ git init
            #     $ git rev-parse HEAD
            #     fatal: ambiguous argument 'HEAD': unknown revision or path not in the working tree.
            return None

        changes = repo.index.diff(None)
        if changes:
            # Ensure that, no matter what happens, the source date is more recent than the check-in date
            source_date = source_date + timedelta(seconds=1)

        # Ensure a more accurate source date is used if there have been any changes to the tracked sources
        for diff in changes:
            if diff.deleted_file:
                continue

            try:
                st = os.lstat(os.path.join(repo.working_dir, diff.b_path))
            except OSError:
                pass
            else:
                file_date = datetime.utcfromtimestamp(st.st_mtime).replace(tzinfo=tzlocal())
                source_date = max(source_date, file_date)

        log.debug("Date of last modification to source: %s", source_date)
        return source_date
    finally:
        if not isinstance(workspace, git.Repo):
            repo.close()


_max_git_objects = 100e6
_git_abbrev_len = math.ceil(math.log2(_max_git_objects) / 2)


def determine_git_version(repo: git.Repo) -> GitVersion:
    """
    Determines the current version of a git repository based on its tags.
    """

    return GitVersion.from_description(
        repo.git.describe(tags=True, long=True, dirty=True, always=True, abbrev=_git_abbrev_len),
    )


def determine_version(
    version_info: Mapping[str, Any],
    config_dir: Optional[PathLike],
    code_dir: Optional[PathLike] = None,
) -> Tuple[Optional[Version], Optional[str]]:
    """
    Determines the current version for the given version configuration snippet.
    """

    version = commit_hash = None

    if 'file' in version_info:
        params = {}
        if 'format' in version_info:
            params['format'] = version_info['format']
        assert config_dir is not None
        fname = os.path.join(config_dir, version_info['file'])
        if os.path.isfile(fname):
            version = read_version(fname, **params)

    if code_dir is not None:
        try:
            with git.Repo(code_dir) as repo:
                gitversion = determine_git_version(repo)
                commit_hash = gitversion.commit_hash

                if version_info.get('tag', False) and version is None:
                    params = {}
                    if 'format' in version_info:
                        params['format'] = version_info['format']
                    if gitversion.dirty:
                        params['dirty_date'] = determine_source_date(repo)

                    version = gitversion.to_version(**params)
                    if version is None and log.isEnabledFor(logging.WARNING):
                        log.warning(
                            "Failed to determine version for %s from tags:\n%s",
                            repo.head.commit,
                            "\n".join(f"    {tag.name}={tag.object.hexsha}^{{}}={tag.commit.hexsha}" for tag in repo.tags),
                        )
        except (git.InvalidGitRepositoryError, git.NoSuchPathError, git.CommandError):
            pass

    return version, commit_hash


@unique
class GitObjectType(Enum):
    regular_file = 0b1000
    symlink = 0b1010
    gitlink = 0b1110


def determine_mtime_from_git(
    repo: git.Repo,
    files: Optional[Iterable[Union[bytes, str]]] = None,
    author_time: bool = False,
) -> Iterable[Tuple[PathLike, GitObjectType, int]]:
    encoding = sys.getfilesystemencoding() or sys.getdefaultencoding()

    if files is None:
        files = set(filter(None, repo.git.ls_files('-z', stdout_as_string=False).split(b'\0')))
    else:
        files = set((fname.encode(encoding) if isinstance(fname, str) else fname) for fname in files)

    log.debug('restoring mtime from git')

    # Set all files' modification times to their last commit's time
    whatchanged = repo.git.whatchanged(pretty='format:%at' if author_time else 'format:%ct', as_process=True)
    mtime = 0
    for line in whatchanged.stdout:
        if not files:
            break

        line = line.strip()
        if not line:
            continue
        if line.startswith(b':'):
            line = line[1:]

            props, filenames = line.split(b'\t', 1)
            old_mode, new_mode, old_hash, new_hash, operation = props.split(b' ')
            old_mode, new_mode = int(old_mode, 8), int(new_mode, 8)

            object_type = (new_mode >> (9 + 3)) & 0b1111
            try:
                object_type = GitObjectType(object_type)
            except ValueError:
                pass

            filenames = filenames.split(b'\t')
            if len(filenames) == 1:
                filenames.insert(0, None)
            old_filename, new_filename = filenames

            if new_filename in files:
                files.remove(new_filename)
                yield new_filename.decode(encoding), object_type, mtime
        else:
            mtime = int(line)
    try:
        whatchanged.terminate()
    except OSError:
        pass


def restore_mtime_from_git(repo: git.Repo, files: Optional[Iterable[Union[bytes, str]]] = None) -> None:
    for filename, object_type, mtime in determine_mtime_from_git(repo, files):
        path = os.path.join(repo.working_tree_dir, filename)
        if object_type == GitObjectType.symlink:
            # Only attempt to modify symlinks' timestamps when the current system supports it.
            # E.g. Python >= 3.3 and Linux kernel >= 2.6.22
            if os.utime in getattr(os, 'supports_follow_symlinks', set()):
                os.utime(path, (mtime, mtime), follow_symlinks=False)
        elif object_type == GitObjectType.gitlink:
            # Skip gitlinks: used by submodules, they don't exist as regular files
            pass
        elif object_type == GitObjectType.regular_file:
            os.utime(path, (mtime, mtime))



# Copyright (c) 2018 - 2021 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

from .utils import (
    determine_config_file_name,
)

from ..config_reader import (
        expand_vars,
        read as read_config,
    )


def _option_from_args(args, option):
    try:
        return args[args.index(option) + 1]
    except Exception:
        for arg in args:
            if arg.startswith(option + '='):
                return arg[len(option + '='):]


def _config_from_args(args):
    workspace = _option_from_args(args, "--workspace")
    if workspace is not None:
        workspace = Path(workspace)
    config = _option_from_args(args, "--config")
    if config is not None:
        config = Path(expand_vars(os.environ, config)).expanduser()
        if workspace is None:
            workspace = config.parent
    else:
        if workspace is None:
            workspace = Path.cwd()
        config = determine_config_file_name(None, workspace=workspace)
    return read_config(config, {'WORKSPACE': str(workspace)})


def phase_from_config(ctx, args, incomplete):
    try:
        cfg = _config_from_args(args)
        for phase in cfg['phases']:
            if incomplete in phase:
                yield phase
    except Exception:
        pass


def variant_from_config(ctx, args, incomplete):
    try:
        cfg = _config_from_args(args)
        phase = _option_from_args(args, '--phase')

        seen_variants = set()
        for phasename, curphase in cfg['phases'].items():
            if phase is not None and phasename != phase:
                continue
            for variant in curphase:
                if variant in seen_variants:
                    continue
                seen_variants.add(variant)
                yield variant
    except Exception:
        pass


def modality_from_config(ctx, args, incomplete):
    try:
        cfg = _config_from_args(args)
        for modality in cfg['modality-source-preparation']:
            if incomplete in modality:
                yield modality
    except Exception:
        pass


def click_log_verbosity(ctx, args, incomplete):
    for level in (
                'DEBUG',
                'INFO',
                'WARNING',
                'ERROR',
                'CRITICAL',
            ):
        if incomplete in level:
            yield level



# Copyright (c) 2020 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from io import StringIO

from ..compat import metadata

PACKAGE : str = __package__.split('.')[0]

# Converting to frozenset because this returns two duplicate entries since importlib_metadata 4.3.0.
# Specifically it seems that the f"{PACKAGE}.egg-info" directory created by setup.py in the repository's root directory gets selected along with the
# f"{PACKAGE}-{VERSION}.dist-info" directory in Tox' virtualenv.
(_hopic_ep,) = frozenset(metadata.entry_points(group="console_scripts", name=PACKAGE))
hopic_cli = _hopic_ep.load()
source_date_epoch = 7 * 24 * 3600

sgr_re = re.compile(r"\x1B\[.*?m")


def config_file(name: str, content: str):
    f = StringIO(content)
    f.name = name
    return f
